# Copyright 2013-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



import urllib.parse

from reahl.tofu import expected
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import Browser, XPath
from reahl.component.modelinterface import Event, Field, Action, ExposedNames, IntegerField
from reahl.component.exceptions import ProgrammerError

from reahl.web.fw import UserInterface, ViewPreCondition, Redirect, Detour, Return, IdentityDictionary, UrlBoundView
from reahl.web.ui import HTML5Page, Form, ButtonInput, A

from reahl.web_dev.fixtures import WebFixture, BasicPageLayout


class FormWithButton(Form):
    def __init__(self, view, event):
        super().__init__(view, 'test_events')
        self.add_child(ButtonInput(self, event))
                

@with_fixtures(WebFixture)
def test_basic_transition(web_fixture):
    """Transitions express how the browser is ferried between Views in reaction to user-initiated Events."""
    fixture  = web_fixture

    def do_something():
        fixture.did_something = True

    class UIWithTwoViews(UserInterface):
        def assemble(self):
            event = Event(label='Click me', action=Action(do_something))
            event.bind('anevent', None)
            viewa = self.define_view('/viewa', title='View a')
            viewa.set_slot('main', FormWithButton.factory(event))
            viewb = self.define_view('/viewb', title='View b')
            viewb.set_slot('main', FormWithButton.factory(event))
            self.define_transition(event, viewa, viewb)

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            self.define_user_interface('/a_ui',  UIWithTwoViews,  IdentityDictionary(), name='test_ui')

    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    # The transition works from viewa
    fixture.did_something = False
    browser.open('/a_ui/viewa')
    browser.click('//input[@value="Click me"]')
    assert browser.current_url.path == '/a_ui/viewb'
    assert fixture.did_something

    # The transition does not work from viewb
    fixture.did_something = False
    browser.open('/a_ui/viewb')
    with expected(ProgrammerError):
        browser.click('//input[@value="Click me"]')
    assert not fixture.did_something


@with_fixtures(WebFixture)
def test_guards(web_fixture):
    """Guards can be set on Transitions. A Transition is only elegible for firing if its guard is True."""
    fixture = web_fixture
    fixture.guard_value = None
    adjustable_guard = Action(lambda:fixture.guard_value)

    false_guard = Action(lambda:False)

    class UIWithGuardedTransitions(UserInterface):
        def assemble(self):
            event = Event(label='Click me')
            event.bind('anevent', None)
            viewa = self.define_view('/viewa', title='View a')
            viewa.set_slot('main', FormWithButton.factory(event))
            viewb = self.define_view('/viewb', title='View b')
            viewc = self.define_view('/viewc', title='View c')
            self.define_transition(event, viewa, viewb, guard=false_guard)
            self.define_transition(event, viewa, viewc, guard=adjustable_guard)

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            self.define_user_interface('/a_ui',  UIWithGuardedTransitions,  IdentityDictionary(), name='test_ui')


    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    # The transition with True guard is the one followed
    fixture.guard_value = True
    browser.open('/a_ui/viewa')
    browser.click('//input[@value="Click me"]')
    assert browser.current_url.path == '/a_ui/viewc'

    # If there is no Transition with a True guard, fail
    fixture.guard_value = False
    browser.open('/a_ui/viewa')
    with expected(ProgrammerError):
        browser.click('//input[@value="Click me"]')


@with_fixtures(WebFixture)
def test_local_transition(web_fixture):
    """A local Transition has its source as its target."""
    fixture = web_fixture
    def do_something():
        fixture.did_something = True

    fixture.guard_passes = True
    guard = Action(lambda:fixture.guard_passes)

    class UIWithAView(UserInterface):
        def assemble(self):
            event = Event(label='Click me', action=Action(do_something))
            event.bind('anevent', None)
            viewa = self.define_view('/viewa', title='View a')
            viewa.set_slot('main', FormWithButton.factory(event))
            self.define_local_transition(event, viewa, guard=guard)

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            self.define_user_interface('/a_ui',  UIWithAView,  IdentityDictionary(), name='test_ui')


    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    # The transition works from viewa
    fixture.did_something = False
    browser.open('/a_ui/viewa')
    browser.click('//input[@value="Click me"]')
    assert browser.current_url.path == '/a_ui/viewa'
    assert fixture.did_something

    # But it is also guarded
    fixture.guard_passes = False
    browser.open('/a_ui/viewa')
    with expected(ProgrammerError):
        browser.click('//input[@value="Click me"]')


@with_fixtures(WebFixture)
def test_transitions_to_parameterised_views(web_fixture):
    """When a Button is placed for an Event that may trigger a Transition to a parameterised View,
       the Event should be bound to the arguments to be used for the target View, using .with_arguments()"""

    class ModelObject:
        events = ExposedNames()
        events.an_event = lambda i: Event(label='click me', event_argument1=IntegerField(),
                                          event_argument2=IntegerField())

    model_object = ModelObject()

    class MyForm(Form):
        def __init__(self, view, name):
            super().__init__(view, name)
            event = model_object.events.an_event.with_arguments(event_argument1=1, event_argument2=2, view_argument=3)
            self.add_child(ButtonInput(self, event))

    class ParameterisedView(UrlBoundView):
        def assemble(self, event_argument1=None, view_argument=None):
            self.title = 'View with event_argument1: %s%s and view_argument: %s%s' \
                        % (event_argument1, type(event_argument1), view_argument, type(view_argument))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            home = self.define_view('/', title='Home page')

            other_view = self.define_view('/page2', title='Page 2',
                                          view_class=ParameterisedView,
                                          event_argument1=IntegerField(required=True),
                                          view_argument=IntegerField(required=True))
            home.set_slot('main', MyForm.factory('myform'))

            self.define_transition(model_object.events.an_event, home, other_view)


    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    web_fixture.reahl_server.set_app(wsgi_app)
    web_fixture.driver_browser.open('/')

    # when the Action is executed, the correct arguments are passed to the View
    web_fixture.driver_browser.click("//input[@value='click me']")
    assert web_fixture.driver_browser.title == 'View with event_argument1: 1%s and view_argument: 3%s' % (type(1), type(3))


@with_fixtures(WebFixture)
def test_transitions_to_parameterised_views_error(web_fixture):
    """If an Event triggers a Transition to a parameterised View, and it was not bound to the arguments
       expected by the target View, an error is raised."""

    class ModelObject:
        events = ExposedNames()
        events.an_event = lambda i: Event(label='Click me')

    model_object = ModelObject()

    class FormWithIncorrectButtonToParameterisedView(Form):
        def __init__(self, view):
            super().__init__(view, 'test_events')
            self.add_child(ButtonInput(self, model_object.events.an_event.with_arguments(arg1='1', arg2='2')))

    class ParameterisedView(UrlBoundView):
        def assemble(self, object_key=None):
            self.title = 'View for: %s' % object_key

    class UIWithParameterisedViews(UserInterface):
        def assemble(self):
            normal_view = self.define_view('/static', title='Static')
            normal_view.set_slot('main', FormWithIncorrectButtonToParameterisedView.factory())
            parameterised_view = self.define_view('/dynamic', view_class=ParameterisedView, object_key=Field(required=True))
            self.define_transition(model_object.events.an_event, normal_view, parameterised_view)

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            self.define_user_interface('/a_ui',  UIWithParameterisedViews,  IdentityDictionary(), name='test_ui')


    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/a_ui/static')
    with expected(ProgrammerError):
        browser.click(XPath.button_labelled('Click me'))


@with_fixtures(WebFixture)
def test_view_preconditions(web_fixture):
    """Views can have Conditions attached to them - if a ViewPreCondition fails upon a GET or HEAD request,
       the specified exception is raised. For all other requests, HTTPNotFound is raised."""
    class SomeException(Exception):
        pass

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            view = self.define_view('/', title='Hello')
            view.set_slot('main', Form.factory('the_form'))
            failing_precondition = ViewPreCondition(lambda: False, exception=SomeException)
            passing_precondition = ViewPreCondition(lambda: True)
            view.add_precondition(passing_precondition)
            view.add_precondition(failing_precondition)

    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    with expected(SomeException):
        browser.open('/')

    browser.post('/_the_form', {}, status=404)


@with_fixtures(WebFixture)
def test_inverse_view_preconditions(web_fixture):
    """A ViewPreCondition can give you another ViewPreCondition which it itself negated, optionally with its own exception."""
    class SomeException(Exception):
        pass

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page)
            view = self.define_view('/', title='Hello')
            passing_precondition = ViewPreCondition(lambda: True)
            failing_precondition = passing_precondition.negated(exception=SomeException)
            view.add_precondition(failing_precondition)


    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    with expected(SomeException):
        browser.open('/')


@with_fixtures(WebFixture)
def test_redirect(web_fixture):
    """Redirect is a special exception that will redirect the browser to another View."""
    class UIWithRedirect(UserInterface):
        def assemble(self):
            viewa = self.define_view('/viewa', title='A')
            viewb = self.define_view('/viewb', title='B')
            failing_precondition = ViewPreCondition(lambda: False, exception=Redirect(viewb.as_bookmark(self)))
            viewa.add_precondition(failing_precondition)

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page)
            self.define_user_interface('/a_ui',  UIWithRedirect,  IdentityDictionary(), name='test_ui')


    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/a_ui/viewa')
    assert browser.current_url.path == '/a_ui/viewb'


@with_fixtures(WebFixture)
def test_detours_and_return_transitions(web_fixture):
    """Detour is a special exception that will redirect the browser to another View, but it also does the
       necessary housekeeping that will allow a return_transition to let the browser return to where the
       Detour was thrown."""
    fixture = web_fixture

    fixture.make_precondition_pass = False
    class UIWithDetour(UserInterface):
        def assemble(self):
            event = Event(label='Click me')
            event.bind('anevent', None)
            viewa = self.define_view('/viewa', title='View a')

            step1 = self.define_view('/firstStepOfDetour', title='Step 1')
            step1.set_slot('main', FormWithButton.factory(event))
            step2 = self.define_view('/lastStepOfDetour', title='Step 2')
            step2.set_slot('main', FormWithButton.factory(event))

            viewa.add_precondition(ViewPreCondition(lambda: fixture.make_precondition_pass, exception=Detour(step1.as_bookmark(self))))
            self.define_transition(event, step1, step2)
            self.define_return_transition(event, step2)

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            self.define_user_interface('/a_ui',  UIWithDetour,  IdentityDictionary(), name='test_ui')


    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)
    fixture.did_something = False

    fixture.make_precondition_pass = False
    browser.open('/a_ui/viewa')
    assert browser.current_url.path == '/a_ui/firstStepOfDetour'

    browser.click('//input[@type="submit"]')
    assert browser.current_url.path == '/a_ui/lastStepOfDetour'

    fixture.make_precondition_pass = True
    browser.click('//input[@type="submit"]')
    assert browser.current_url.path == '/a_ui/viewa'

    # The query string is cleared after such a return (it is used to remember where to return to)
    assert browser.current_url.query == ''


@with_fixtures(WebFixture)
def test_detours_and_explicit_return_view(web_fixture):
    """A Detour can also explicitly set the View to return to."""

    class UIWithDetour(UserInterface):
        def assemble(self):
            event = Event(label='Click me')
            event.bind('anevent', None)
            viewa = self.define_view('/viewa', title='View a')
            explicit_return_view = self.define_view('/explicitReturnView', title='Explicit Return View')

            detour = self.define_view('/detour', title='Detour')
            detour.set_slot('main', FormWithButton.factory(event))

            viewa.add_precondition(ViewPreCondition(lambda: False, exception=Detour(detour.as_bookmark(self), return_to=explicit_return_view.as_bookmark(self))))
            self.define_return_transition(event, detour)

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            self.define_user_interface('/a_ui',  UIWithDetour,  IdentityDictionary(), name='test_ui')


    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/a_ui/viewa')
    assert browser.current_url.path == '/a_ui/detour'

    browser.click('//input[@type="submit"]')
    assert browser.current_url.path == '/a_ui/explicitReturnView'

    # The query string is cleared after such a return (it is used to remember where to return to)
    assert browser.current_url.query == ''


@with_fixtures(WebFixture)
def test_redirect_used_to_return(web_fixture):
    """A Return is an exception used with Preconditoins to return automatically to another View (as set by detour),
       instead of using a return_transition (the latter can only be triggered by a user)."""

    class UIWithDetour(UserInterface):
        def assemble(self):
            viewa = self.define_view('/viewa', title='View a')
            explicit_return_view = self.define_view('/explicitReturnView', title='Explicit Return View')
            default = self.define_view('/defaultReturnView', title='Default view to return to')
            detour = self.define_view('/detour', title='Detour')

            viewa.add_precondition(ViewPreCondition(lambda: False, exception=Detour(detour.as_bookmark(self), return_to=explicit_return_view.as_bookmark(self))))
            detour.add_precondition(ViewPreCondition(lambda: False, exception=Return(default.as_bookmark(self))))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page)
            self.define_user_interface('/a_ui',  UIWithDetour,  IdentityDictionary(), name='test_ui')


    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    # Normal operation - when a caller can be determined
    browser.open('/a_ui/viewa')
    assert browser.current_url.path == '/a_ui/explicitReturnView'

    #  - the query string is cleared after such a return (it is used to remember where to return to)
    assert browser.current_url.query == ''

    # When a caller cannot be determined, the default is used
    browser.open('/a_ui/detour')
    assert browser.current_url.path == '/a_ui/defaultReturnView'

    #  - the query string is cleared after such a return (it is used to remember where to return to)
    assert browser.current_url.query == ''


@with_fixtures(WebFixture)
def test_unconditional_redirection(web_fixture):
    """You can force an URL to always redirect to a given Bookmark."""

    class UIWithRedirect(UserInterface):
        def assemble(self):
            self.define_view('/target', title='')
            self.define_redirect('/redirected', self.get_bookmark(relative_path='/target'))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page)
            self.define_user_interface('/a_ui',  UIWithRedirect,  IdentityDictionary(), name='test_ui')


    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/a_ui/redirected')
    assert browser.current_url.path == '/a_ui/target'


@with_fixtures(WebFixture)
def test_linking_to_views_marked_as_detour(web_fixture):
    """A View can be marked as the start of a Detour. Where used, a Bookmark for such a View
       will automatically include a returnTo in the its query string. This allows an
       eventual return_transition (or similar) to return to where, eg, a link was clicked from.
       This mechanism works for returning across UserInterfaces."""

    class UIWithLink(UserInterface):
        def assemble(self, bookmark=None):
            self.bookmark = bookmark
            self.define_view('/initial', title='View a').set_slot('main', A.factory_from_bookmark(self.bookmark))


    class UIWithDetour(UserInterface):
        def assemble(self):
            event = Event(label='Click me')
            event.bind('anevent', None)

            step1 = self.define_view('/firstStepOfDetour', title='Step 1', detour=True)
            step1.set_slot('main', FormWithButton.factory(event))
            step2 = self.define_view('/lastStepOfDetour', title='Step 2')
            step2.set_slot('main', FormWithButton.factory(event))

            self.define_transition(event, step1, step2)
            self.define_return_transition(event, step2)

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            detour_ui = self.define_user_interface('/uiWithDetour',  UIWithDetour,  IdentityDictionary(), name='second_ui')
            bookmark = detour_ui.get_bookmark(relative_path='/firstStepOfDetour')
            self.define_user_interface('/uiWithLink',  UIWithLink,  IdentityDictionary(), name='first_ui', bookmark=bookmark)


    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/uiWithLink/initial')
    browser.click('//a')
    assert browser.current_url.path == '/uiWithDetour/firstStepOfDetour'

    browser.click('//input[@type="submit"]')
    assert browser.current_url.path == '/uiWithDetour/lastStepOfDetour'

    browser.click('//input[@type="submit"]')
    assert browser.current_url.path == '/uiWithLink/initial'

    # The query string is cleared after such a return (it is used to remember where to return to)
    assert browser.current_url.query == ''


@with_fixtures(WebFixture)
def test_detour_is_non_reentrant(web_fixture):
    """Once detoured to a View marked as the start of a Detour, a Bookmark to that View itself
    will not re-enter the detour.
    """

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())

            step1 = self.define_view('/firstStepOfDetour', title='Step 1', detour=True)
            step1.set_slot('main', A.factory_from_bookmark(step1.as_bookmark(self)))

            home = self.define_view('/initial', title='View a')
            home.set_slot('main', A.factory_from_bookmark(step1.as_bookmark(self)))


    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    def locationIsSetToReturnTo(url_path):
        return urllib.parse.parse_qs(browser.current_url.query)['returnTo'] == [url_path]

    browser.open('/initial')
    browser.click(XPath.link().with_text('Step 1'))
    assert browser.current_url.path == '/firstStepOfDetour'
    assert locationIsSetToReturnTo('http://localhost/initial')

    browser.click(XPath.link().with_text('Step 1'))
    assert browser.current_url.path == '/firstStepOfDetour'
    assert locationIsSetToReturnTo('http://localhost/initial')



# Copyright 2018-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



import threading

from sqlalchemy import Column, Integer

from reahl.tofu import Fixture, scenario, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.web_dev.fixtures import WebFixture
from reahl.browsertools.browsertools import XPath
from reahl.web.fw import Widget, Url
from reahl.web.ui import Form, Div, SelectInput, Label, P, RadioButtonSelectInput, CheckboxSelectInput, \
    CheckboxInput, ButtonInput, TextInput, FormLayout, A
from reahl.component.modelinterface import Field, BooleanField, MultiChoiceField, ChoiceField, Choice, ExposedNames, \
    IntegerField, EmailField, Event, Action, Allowed
from reahl.component.exceptions import DomainException
from reahl.web_dev.inputandvalidation.test_widgetqueryargs import QueryStringFixture
from reahl.sqlalchemysupport import Base, Session
from reahl.sqlalchemysupport_dev.fixtures import SqlAlchemyFixture


@uses(web_fixture=WebFixture)
class ResponsiveDisclosureFixture(Fixture):

    def new_ModelObject(self):
        class ModelObject:
            def __init__(self):
                self.choice = 1

            fields = ExposedNames()
            fields.choice = lambda i: ChoiceField([Choice(1, IntegerField(label='One')),
                                                   Choice(2, IntegerField(label='Two')),
                                                   Choice(3, IntegerField(label='Three'))],
                                                  label='Choice')
        return ModelObject

    def new_MainWidget(self):
        fixture = self
        class MainWidget(Widget):
            def __init__(self, view):
                super().__init__(view)
                an_object = fixture.ModelObject()
                self.add_child(fixture.MyForm(view, an_object))

        return MainWidget

    def new_MyForm(self):
        fixture = self
        class MyForm(Form):
            def __init__(self, view, an_object):
                super().__init__(view, 'myform')
                self.enable_refresh()
                self.use_layout(FormLayout())
                self.layout.add_input(fixture.create_trigger_input(self, an_object))
                self.add_child(P(self.view, text='My state is now %s' % an_object.choice))

        return MyForm

    def create_trigger_input(self, form, an_object):
        return SelectInput(form, an_object.fields.choice, refresh_widget=form)



class ResponsiveWidgetScenarios(ResponsiveDisclosureFixture):
    @scenario
    def select_input(self):
        self.expected_focussed_element = XPath.select_labelled('Choice')
        def change_value(browser):
            browser.select(self.expected_focussed_element, 'Three')
        self.change_value = change_value
        self.initial_state = 1
        self.changed_state = 3

    @scenario
    def radio_buttons(self):
        def create_trigger_input(form, an_object):
            return RadioButtonSelectInput(form, an_object.fields.choice, refresh_widget=form)
        self.create_trigger_input = create_trigger_input

        self.expected_focussed_element = XPath.input_labelled('Three')
        def change_value(browser):
            browser.click(self.expected_focussed_element)
        self.change_value = change_value
        self.initial_state = 1
        self.changed_state = 3

    @scenario
    def single_valued_checkbox(self):
        class ModelObject:
            def __init__(self):
                self.choice = False
            fields = ExposedNames()
            fields.choice = lambda i: BooleanField(label=u'Choice',
                                                   true_value=u'true_value', false_value=u'false_value')
        self.ModelObject = ModelObject

        def create_trigger_input(form, an_object):
            the_input = CheckboxInput(form, an_object.fields.choice, refresh_widget=form)
            the_input.set_id('marvin')
            return the_input
        self.create_trigger_input = create_trigger_input

        self.expected_focussed_element = XPath.input_labelled('Choice')
        def change_value(browser):
            browser.click(self.expected_focussed_element)
        self.change_value = change_value
        self.initial_state = False
        self.changed_state = True

    @scenario
    def multi_valued_checkbox_select(self):
        class ModelObject:
            def __init__(self):
                self.choice = [1]

            fields = ExposedNames()
            fields.choice = lambda i: MultiChoiceField([Choice(1, IntegerField(label='One')),
                                                        Choice(2, IntegerField(label='Two')),
                                                        Choice(3, IntegerField(label='Three'))],
                                                       label='Choice')
        self.ModelObject = ModelObject

        def create_trigger_input(form, an_object):
            the_input = CheckboxSelectInput(form, an_object.fields.choice, refresh_widget=form)
            the_input.set_id('marvin')
            return the_input
        self.create_trigger_input = create_trigger_input

        self.expected_focussed_element = XPath.input_labelled('Three')
        def change_value(browser):
            browser.click(self.expected_focussed_element)
        self.change_value = change_value
        self.initial_state = [1]
        self.changed_state = [1, 3]

    @scenario
    def multi_valued_checkbox_select_with_single_choice_corner_case_empty_a_list(self):
        self.multi_valued_checkbox_select()
        class ModelObject:
            def __init__(self):
                self.choice = [1]

            fields = ExposedNames()
            fields.choice = lambda i: MultiChoiceField([Choice(1, IntegerField(label='One'))],
                                                       label='Choice')
        self.ModelObject = ModelObject

        self.expected_focussed_element = XPath.input_labelled('One')
        def change_value(browser):
            browser.click(self.expected_focussed_element)
        self.change_value = change_value
        self.initial_state = [1]
        self.changed_state = []

    @scenario
    def multi_valued_checkbox_select_with_single_choice_corner_case_add_to_empty_list(self):
        self.multi_valued_checkbox_select()
        class ModelObject:
            def __init__(self):
                self.choice = []

            fields = ExposedNames()
            fields.choice = lambda i: MultiChoiceField([Choice(1, IntegerField(label='One'))],
                                                       label='Choice')
        self.ModelObject = ModelObject
        #self.MyForm from multi_valued_checkbox_select

        self.expected_focussed_element = XPath.input_labelled('One')
        def change_value(browser):
            browser.click(self.expected_focussed_element)
        self.change_value = change_value
        self.initial_state = []
        self.changed_state = [1]

    @scenario
    def multi_valued_select(self):
        class ModelObject:
            def __init__(self):
                self.choice = [1]

            fields = ExposedNames()
            fields.choice = lambda i: MultiChoiceField([Choice(1, IntegerField(label='One')),
                                                        Choice(2, IntegerField(label='Two')),
                                                        Choice(3, IntegerField(label='Three'))],
                                                       label='Choice')
        self.ModelObject = ModelObject

        def create_trigger_input(form, an_object):
            the_input = SelectInput(form, an_object.fields.choice, refresh_widget=form)
            the_input.set_id('marvin')
            return the_input
        self.create_trigger_input = create_trigger_input

        self.expected_focussed_element = XPath.select_labelled('Choice')
        def change_value(browser):
            browser.select(self.expected_focussed_element, 'Three')
        self.change_value = change_value
        self.initial_state = [1]
        self.changed_state = [1, 3]


@with_fixtures(WebFixture, QueryStringFixture, ResponsiveWidgetScenarios)
def test_inputs_can_refresh_parent_widgets(web_fixture, query_string_fixture, responsive_widget_scenarios):
    """An Input can cause another Widget to be re-rendered if the input value changes."""

    fixture = responsive_widget_scenarios

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MainWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    assert browser.wait_for(query_string_fixture.is_state_now, fixture.initial_state)
    fixture.change_value(browser)
    assert browser.wait_for(query_string_fixture.is_state_now, fixture.changed_state)


@with_fixtures(WebFixture, QueryStringFixture, ResponsiveDisclosureFixture)
def test_refresh_input_can_be_set_after_input_is_created(web_fixture, query_string_fixture, responsive_disclosure_fixture):
    """`set_refresh_widgets` can be called after construction of the Input to instruct it to refresh a given list of Widgets."""
    fixture = responsive_disclosure_fixture

    class MyForm(Form):
        def __init__(self, view, an_object):
            super().__init__(view, 'myform')
            self.use_layout(FormLayout())
            trigger_input = self.layout.add_input(SelectInput(self, an_object.fields.choice))
            refreshed_widget = self.add_child(P(self.view, text='My state is now %s' % an_object.choice))
            refreshed_widget.set_id('refreshing-paragraph')
            refreshed_widget.enable_refresh()
            
            another_widget_to_refresh = self.add_child(P(self.view, text='Another widget state is now %s' % an_object.choice))
            another_widget_to_refresh.set_id('another-refreshing-paragraph')
            another_widget_to_refresh.enable_refresh()
            
            trigger_input.set_refresh_widgets([refreshed_widget, another_widget_to_refresh])

    fixture.MyForm = MyForm

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MainWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    assert browser.wait_for(query_string_fixture.is_state_now, 1)
    assert browser.wait_for(query_string_fixture.is_state_on_another_widget_now, 1)
    browser.select(XPath.select_labelled('Choice'), 'Three')
    assert browser.wait_for(query_string_fixture.is_state_now, 3)
    assert browser.wait_for(query_string_fixture.is_state_on_another_widget_now, 3)


@with_fixtures(WebFixture, QueryStringFixture, ResponsiveDisclosureFixture)
def test_refresh_widgets_with_widget_arguments(web_fixture, query_string_fixture, responsive_disclosure_fixture):
    """Widget arguemnts are taken into account when widgets are refreshed by an input"""
    fixture = responsive_disclosure_fixture

    class MyForm(Form):
        def __init__(self, view, an_object):
            self.arg = 'default'
            super().__init__(view, 'myform')
            self.enable_refresh()
            self.use_layout(FormLayout())
            trigger_input = self.layout.add_input(SelectInput(self, an_object.fields.choice))

            refreshed_widget = self.add_child(P(self.view, text='%s My state is now %s' % (self.arg, an_object.choice)))
            refreshed_widget.set_id('refreshing-paragraph1')
            refreshed_widget.enable_refresh()

            trigger_input.set_refresh_widgets([refreshed_widget])

            link = self.add_child(A(view, Url('#arg=changed'), description='Hash change link'))

        query_fields = ExposedNames()
        query_fields.arg = lambda i: Field()

    fixture.MyForm = MyForm

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MainWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    assert browser.wait_for(query_string_fixture.is_state_labelled_now, 'default My state',  1)
    browser.select(XPath.select_labelled('Choice'), 'Three')
    assert browser.wait_for(query_string_fixture.is_state_labelled_now, 'default My state',  3)

    browser.click(XPath.link().with_text('Hash change link'))
    assert browser.wait_for(query_string_fixture.is_state_labelled_now, 'changed My state',  3)
    browser.select(XPath.select_labelled('Choice'), 'Two')
    assert browser.wait_for(query_string_fixture.is_state_labelled_now, 'changed My state',  2)


@with_fixtures(WebFixture, QueryStringFixture, ResponsiveDisclosureFixture)
def test_handle_ajax_error(web_fixture, query_string_fixture, responsive_disclosure_fixture):
    """."""
    fixture = responsive_disclosure_fixture

    class MyForm(fixture.MyForm):
        def __init__(self, view, an_object):
            super().__init__(view, an_object)
            if an_object.choice == 2:
                raise Exception('Intentionally breaking')

    fixture.MyForm = MyForm

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MainWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser

    wsgi_app.config.reahlsystem.debug = False #causes redirect to the error page and not stacktrace

    browser.open('/')

    assert browser.wait_for(query_string_fixture.is_state_now, 1)
    browser.select(XPath.select_labelled('Choice'), 'Two')
    assert browser.wait_for_element_visible(XPath.heading(1).with_text('An error occurred:'))
    assert browser.wait_for_element_visible(XPath.paragraph().including_text('Intentionally breaking'))


@with_fixtures(WebFixture, QueryStringFixture, ResponsiveDisclosureFixture)
def test_overridden_names(web_fixture, query_string_fixture, responsive_disclosure_fixture):
    """The overridden names of inputs correctly ensures that that input's state is distinguished from another with the same name."""
    fixture = responsive_disclosure_fixture

    class ModelObject:
        def __init__(self):
            self.choice = [1]

        fields = ExposedNames()
        fields.choice = lambda i: MultiChoiceField([Choice(1, IntegerField(label='One')),
                                                    Choice(2, IntegerField(label='Two')),
                                                    Choice(3, IntegerField(label='Three'))],
                                                   label='Choice')
    fixture.ModelObject = ModelObject

    def create_trigger_input(form, an_object):
        the_input = CheckboxSelectInput(form, an_object.fields.choice.with_discriminator('first'), refresh_widget=form)
        the_input.set_id('marvin')
        return the_input
    fixture.create_trigger_input = create_trigger_input

    class MyForm(fixture.MyForm):
        def __init__(self, view, an_object):
            super().__init__(view, an_object)
            another_model_object = fixture.ModelObject()
            self.layout.add_input(CheckboxSelectInput(self, another_model_object.fields.choice))

    fixture.MyForm = MyForm

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MainWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    browser.click('//input[@name="myform-first-choice[]" and @value="3"]')
    assert browser.wait_for(query_string_fixture.is_state_now, [1,3])


class BlockingRefreshFixture(ResponsiveDisclosureFixture):
    should_pause_to_simulate_long_refresh = False

    def new_block_event(self):
        return threading.Event()

    def simulate_long_refresh_start(self):
        self.block_event.wait()

    def simulate_long_refresh_done(self):
        self.block_event.set()

    def new_MyForm(self):
        fixture = self
        class MyFormThatPauses(super().new_MyForm()):
            def __init__(self, view, model_object):
                super().__init__(view, model_object)
                if fixture.should_pause_to_simulate_long_refresh:
                    fixture.simulate_long_refresh_start()

        return MyFormThatPauses

    def is_form_blocked(self, browser):
        form_xpath = XPath('//form[@id="myform"]')
        form_blocked_xpath = XPath('%s/div[@class="blockUI"]' % form_xpath)
        return browser.is_element_present(form_blocked_xpath)


@with_fixtures(WebFixture, BlockingRefreshFixture, QueryStringFixture)
def test_the_form_is_blocked_while_the_widget_is_being_refreshed(web_fixture, blocking_refresh_fixture, query_string_fixture):
    """The form is blocked until the responsive widget has refreshed it contents."""

    fixture = blocking_refresh_fixture

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MainWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser

    fixture.should_pause_to_simulate_long_refresh = False
    browser.open('/')

    fixture.should_pause_to_simulate_long_refresh = True
    with web_fixture.reahl_server.in_background(wait_till_done_serving=False):
        browser.click(XPath.option().with_text('Three'), wait_for_ajax=False)

    assert fixture.is_form_blocked(browser)

    fixture.simulate_long_refresh_done()

    assert browser.wait_for(query_string_fixture.is_state_now, 3)
    assert not fixture.is_form_blocked(browser)


@with_fixtures(WebFixture, ResponsiveDisclosureFixture, SqlAlchemyFixture, QueryStringFixture)
def test_form_values_are_not_persisted_until_form_is_submitted(web_fixture, responsive_disclosure_fixture, sql_alchemy_fixture, query_string_fixture):
    """Values submitted via ajax are used only to redraw the screen; they are only changed on the underlying domain once the form is submitted."""

    fixture = responsive_disclosure_fixture

    class ModelObject(Base):
        __tablename__ = 'test_responsive_disclosure_rollback'
        id = Column(Integer, primary_key=True)
        choice = Column(Integer, default=1)

        fields = ExposedNames()
        fields.choice = lambda i: ChoiceField([Choice(1, IntegerField(label='One')),
                                               Choice(2, IntegerField(label='Two')),
                                               Choice(3, IntegerField(label='Three'))],
                                              label='Choice')
    fixture.ModelObject = ModelObject

    class FormWithButton(fixture.MyForm):
        def __init__(self, view, an_object):
            super().__init__(view, an_object)

            self.define_event_handler(self.events.submit)
            self.add_child(ButtonInput(self, self.events.submit))

        events = ExposedNames()
        events.submit = lambda i: Event(label='Submit')

    fixture.MyForm = FormWithButton

    with sql_alchemy_fixture.persistent_test_classes(ModelObject):

        fixture.model_object = ModelObject()
        Session.add(fixture.model_object)

        class MainWidgetWithPersistentModelObject(Widget):
            def __init__(self, view):
                super().__init__(view)
                an_object = fixture.model_object
                self.add_child(fixture.MyForm(view, an_object))

        wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=MainWidgetWithPersistentModelObject.factory())
        web_fixture.reahl_server.set_app(wsgi_app)
        browser = web_fixture.driver_browser
        browser.open('/')

        assert fixture.model_object.choice == 1
        browser.click(XPath.option().with_text('Three'))

        assert browser.wait_for(query_string_fixture.is_state_now, 3)  # The screen was updated,
        assert fixture.model_object.choice == 1                                # but the database not.

        browser.click(XPath.button_labelled('Submit'))
        assert browser.wait_for(query_string_fixture.is_state_now, 3)
        assert fixture.model_object.choice == 3                                # Now the database is updated too.


class DisclosedInputFixture(Fixture):

    def new_trigger_input_type(self):
        return CheckboxInput

    def new_raise_domain_exception_on_submit(self):
        return False

    def new_default_trigger_field_value(self):
        return True

    def new_MyForm(self):
        fixture = self
        class ModelObject:
            def __init__(self):
                self.trigger_field = fixture.default_trigger_field_value
                self.email = None

            events = ExposedNames()
            events.an_event = lambda i: Event(label='click me', action=Action(i.submit))

            def submit(self):
                if fixture.raise_domain_exception_on_submit:
                    raise DomainException()
                fixture.submitted_model_object = self

            fields = ExposedNames()
            fields.trigger_field = lambda i: BooleanField(label='Trigger field')
            fields.email = lambda i: EmailField(required=True, label='Email')

        class MyForm(Form):
            def __init__(self, view):
                super().__init__(view, 'myform')
                self.enable_refresh()
                self.use_layout(FormLayout())

                if self.exception:
                    self.add_child(P(view, text='Exception raised'))

                model_object = ModelObject()
                self.layout.add_input(fixture.trigger_input_type(self, model_object.fields.trigger_field, refresh_widget=self))

                if model_object.trigger_field:
                    self.layout.add_input(TextInput(self, model_object.fields.email))

                self.define_event_handler(model_object.events.an_event)
                self.add_child(ButtonInput(self, model_object.events.an_event))
        return MyForm


@with_fixtures(WebFixture, DisclosedInputFixture)
def test_validation_of_undisclosed_yet_required_input(web_fixture, disclosed_input_fixture):
    """If a Field has a required constraint, but its Input is not currently displayed as part of the form (because of the
       state of another Input), and the form is submitted, the constraint should not cause an exception(input was omitted)."""

    fixture = disclosed_input_fixture

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MyForm.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    assert browser.is_element_present(XPath.input_labelled('Email'))
    browser.click(XPath.input_labelled('Trigger field'))
    assert not browser.is_element_present(XPath.input_labelled('Email'))
    browser.click(XPath.button_labelled('click me'))

    assert not fixture.submitted_model_object.email


@with_fixtures(WebFixture, DisclosedInputFixture)
def test_input_values_retained_upon_domain_exception(web_fixture, disclosed_input_fixture):
    """When a domain exception occurs the values typed into an input are retained on the error page."""
    # Note: the implementation of this is different for dynamic stuff than for plain old forms.
    #
    # Client state is maintained (for a subsequent GET) by posting it to the server in a hidden input; and also saved on an exception.
    # This state is used at Widget construction stage (before the POST is handled) to ensure all Widgets on the View are built as 
    # before, and as per the client state.
    # 
    # This test is about how that state is managed and when which version of the state (saved in DB, POSTed etc) takes precedence.

    fixture = disclosed_input_fixture

    fixture.raise_domain_exception_on_submit = True
    fixture.default_trigger_field_value = False

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MyForm.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    assert not browser.is_element_present(XPath.input_labelled('Email'))
    browser.click(XPath.input_labelled('Trigger field'))
    browser.type(XPath.input_labelled('Email'), 'expectme@example.org')
    browser.click(XPath.button_labelled('click me'))

    assert browser.is_element_present(XPath.paragraph().including_text('Exception raised'))
    assert browser.is_selected(XPath.input_labelled('Trigger field'))
    assert browser.get_value(XPath.input_labelled('Email')) == 'expectme@example.org'


@with_fixtures(WebFixture, DisclosedInputFixture)
def test_inputs_cleared_after_domain_exception_resubmit(web_fixture, disclosed_input_fixture):
    """After a domain exception followed by a successful submit, saved state on the server is 
       cleared, so that newly rendered inputs on the same URL will have defaulted values again."""
    # Also see related comment on test_input_values_retained_upon_domain_exception

    fixture = disclosed_input_fixture

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MyForm.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser

    # First get a domain exception
    fixture.raise_domain_exception_on_submit = True
    fixture.default_trigger_field_value = False
    browser.open('/')

    browser.click(XPath.input_labelled('Trigger field'))
    browser.type(XPath.input_labelled('Email'), 'expectme@example.org')
    browser.click(XPath.button_labelled('click me'))
    assert browser.is_element_present(XPath.paragraph().including_text('Exception raised'))

    # Then successful commit
    fixture.raise_domain_exception_on_submit = False

    browser.click(XPath.button_labelled('click me'))

    # Values are all defaulted like on a first render
    assert not browser.is_element_present(XPath.paragraph().including_text('Exception raised'))
    assert not browser.is_selected(XPath.input_labelled('Trigger field'))
    assert not browser.is_element_present(XPath.paragraph().including_text('Email'))
    browser.click(XPath.input_labelled('Trigger field'))
    assert browser.get_value(XPath.input_labelled('Email')) == ''


class DirectionScenarios(Fixture):
    @scenario
    def forwards(self):
        self.going_forwards = True

    @scenario
    def backwards(self):
        self.going_forwards = False


class InputOrderScenarios(Fixture):
    @scenario
    def next_input_appears(self):
        self.start_with_other_present = False
        self.add_button = True
        self.expected_next_focussed = XPath.input_labelled('Other')

    @scenario
    def next_input_disappears(self):
        self.start_with_other_present = True
        self.add_button = True
        self.expected_next_focussed = XPath.button_labelled('click me')

    @scenario
    def last_input_disappears(self):
        self.start_with_other_present = True
        self.add_button = False
        self.expected_next_focussed = XPath.input_labelled('Edge')


@uses(direction=DirectionScenarios, input_order=InputOrderScenarios)
class TabOrderFixture(Fixture):

    @property
    def going_forwards(self):
        return self.direction.going_forwards

    @property
    def add_button(self):
        return self.input_order.add_button

    @property
    def start_with_other_present(self):
        return self.input_order.start_with_other_present

    def new_MyForm(self):
        fixture = self
        class ModelObject:
            def __init__(self):
                self.trigger_field = fixture.start_with_other_present
                self.other = None
                self.edge = None

            events = ExposedNames()
            events.an_event = lambda i: Event(label='click me')

            fields = ExposedNames()
            fields.trigger_field = lambda i: BooleanField(label='Trigger field')
            fields.other = lambda i: Field(label='Other')
            fields.edge = lambda i: Field(label='Edge')

        class MyForm(Form):
            def __init__(self, view):
                super().__init__(view, 'myform')
                self.enable_refresh()
                self.use_layout(FormLayout())
                if self.exception:
                    self.add_child(P(view, text='Exception raised'))

                model_object = ModelObject()

                trigger_input = TextInput(self, model_object.fields.trigger_field, refresh_widget=self)
                self.define_event_handler(model_object.events.an_event)

                if fixture.going_forwards:
                    self.layout.add_input(TextInput(self, model_object.fields.edge))
                else:
                    if fixture.add_button:
                        self.add_child(ButtonInput(self, model_object.events.an_event))
                    if model_object.trigger_field:
                        self.layout.add_input(TextInput(self, model_object.fields.other))

                self.layout.add_input(trigger_input)

                if fixture.going_forwards:
                    if model_object.trigger_field:
                        self.layout.add_input(TextInput(self, model_object.fields.other))
                    if fixture.add_button:
                        self.add_child(ButtonInput(self, model_object.events.an_event))
                else:
                    self.layout.add_input(TextInput(self, model_object.fields.edge))
        return MyForm


@with_fixtures(WebFixture, TabOrderFixture)
def test_correct_tab_order_for_responsive_widgets(web_fixture, disclosed_input_trigger_fixture):
    """When a user TABs out of an input that then triggers a change, focus is shifted to the correct element as per the page after the refresh."""

    fixture = disclosed_input_trigger_fixture

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MyForm.factory())

    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    current_value = browser.get_value(XPath.input_labelled('Trigger field'))
    new_value = 'on' if current_value == 'off' else 'off'

    browser.type(XPath.input_labelled('Trigger field'), new_value, trigger_blur=False, wait_for_ajax=False)
    browser.press_tab(shift=not fixture.going_forwards)
    assert browser.wait_for(browser.is_focus_on, fixture.input_order.expected_next_focussed)


@with_fixtures(WebFixture, QueryStringFixture, ResponsiveWidgetScenarios)
def test_focus_location_after_refresh_without_tabbing(web_fixture, query_string_fixture, responsive_widget_scenarios):
    """Check that focus remains on changed widget after refresh, without having to trigger the change by pressing TAB."""

    fixture = responsive_widget_scenarios

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MainWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    assert browser.wait_for(query_string_fixture.is_state_now, fixture.initial_state)
    fixture.change_value(browser)
    assert browser.wait_for(query_string_fixture.is_state_now, fixture.changed_state)
    assert browser.is_focus_on(fixture.expected_focussed_element)


class TimingFixture(BlockingRefreshFixture):
    def new_ModelObject(self):
        class ModelObject:
            def __init__(self):
                self.choice = 1

            fields = ExposedNames()
            fields.some_text = lambda i: Field(label='Some Text')
            fields.trigger_field = lambda i: Field(label='Trigger')
        return ModelObject

    def new_MyForm(self):
        fixture = self
        class MyFormWithTextInput(super().new_MyForm()):
            def __init__(self, view, model_object):
                super().__init__(view, model_object)
                self.layout.add_input(TextInput(self, model_object.fields.some_text))
        return MyFormWithTextInput

    def create_trigger_input(self, form, an_object):
        return TextInput(form, an_object.fields.trigger_field, refresh_widget=form)


@with_fixtures(WebFixture, TimingFixture, QueryStringFixture)
def test_while_refreshing_typed_input_is_discarded(web_fixture, blocking_refresh_fixture, query_string_fixture):
    """When the user tabs out all keyboard input is ignored until after the page has changed."""

    fixture = blocking_refresh_fixture

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MainWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser

    fixture.should_pause_to_simulate_long_refresh = False
    browser.open('/')

    fixture.should_pause_to_simulate_long_refresh = True
    with web_fixture.reahl_server.in_background():
        browser.type(XPath.input_labelled('Trigger'), 'qwerty', trigger_blur=False, wait_for_ajax=False)
        browser.press_tab()
        browser.type_focussed('abc')
        assert 'abc' not in browser.get_value(XPath.input_labelled('Some Text'))
        assert 'abc' not in browser.get_value(XPath.input_labelled('Trigger'))
        fixture.simulate_long_refresh_done()
    
    assert browser.wait_for_not(fixture.is_form_blocked, browser)

    assert 'abc' not in browser.get_value(XPath.input_labelled('Some Text'))
    assert 'abc' not in browser.get_value(XPath.input_labelled('Trigger'))
    browser.type_focussed('asd')
    assert 'asd' in browser.get_value(XPath.input_labelled('Some Text'))


@with_fixtures(WebFixture, DisclosedInputFixture)
def test_ignore_button_click_on_change(web_fixture, disclosed_input_trigger_fixture):
    """If a button click triggers a change to the page (due to a modified TextInput losing focus), the click is ignored."""

    fixture = disclosed_input_trigger_fixture
    fixture.trigger_input_type = TextInput
    fixture.default_trigger_field_value = False

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MyForm.factory())

    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    assert browser.get_value(XPath.input_labelled('Trigger field')) == 'off'
    assert not browser.is_element_present(XPath.input_labelled('Email'))

    browser.type(XPath.input_labelled('Trigger field'), 'on', trigger_blur=False)
    with browser.no_page_load_expected():
        browser.click(XPath.button_labelled('click me'))

    assert browser.is_focus_on(XPath.input_labelled('Trigger field'))
    assert browser.is_element_present(XPath.input_labelled('Email'))
    assert browser.is_on_top(XPath.button_labelled('click me'))


class NestedResponsiveDisclosureFixture(Fixture):

    def new_ModelObject(self):
        class ModelObject:
            def __init__(self):
                self.trigger_field = False
                self.nested_trigger_field = False

            fields = ExposedNames()
            fields.trigger_field = lambda i: BooleanField(label='Trigger field')
            fields.nested_trigger_field = lambda i: BooleanField(label='Nested trigger field')

        return ModelObject

    def new_MyForm(self):
        fixture = self
        class MyNestedChangingWidget(Div):
            def __init__(self, form, model_object):
                self.model_object = model_object
                super().__init__(form.view, css_id='nested_changing_widget')
                self.enable_refresh()

                nested_checkbox_input = CheckboxInput(form, model_object.fields.nested_trigger_field, refresh_widget=self)
                self.add_child(Label(self.view, for_input=nested_checkbox_input))
                self.add_child(nested_checkbox_input)

                if self.model_object.nested_trigger_field:
                    self.add_child(P(self.view, 'My state is now showing nested responsive content'))

        class MyForm(Form):
            def __init__(self, view):
                super().__init__(view, 'myform')
                self.enable_refresh()
                self.use_layout(FormLayout())
                model_object = fixture.ModelObject()

                self.layout.add_input(CheckboxInput(self, model_object.fields.trigger_field, refresh_widget=self))

                if model_object.trigger_field:
                    self.add_child(P(self.view, 'My state is now showing outer responsive content'))
                    self.add_child(MyNestedChangingWidget(self, model_object))


        return MyForm

    def are_all_parts_enabled(self, browser):
        return browser.is_interactable(XPath.input_labelled('Trigger field')) and \
            browser.is_interactable(XPath.input_labelled('Nested trigger field')) and \
            browser.is_on_top(XPath.paragraph().including_text('showing nested responsive content'))


@with_fixtures(WebFixture, QueryStringFixture, NestedResponsiveDisclosureFixture)
def test_inputs_and_widgets_work_when_nested(web_fixture, query_string_fixture, nested_responsive_disclosure_fixture):
    """Refreshable widgets can be nested inside a target widget."""

    fixture = nested_responsive_disclosure_fixture
    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MyForm.factory())

    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    assert browser.wait_for_not(query_string_fixture.is_state_now, 'showing outer responsive content')
    assert browser.wait_for_not(query_string_fixture.is_state_now, 'showing nested responsive content')

    browser.click(XPath.input_labelled('Trigger field'))
    assert browser.wait_for(query_string_fixture.is_state_now, 'showing outer responsive content')
    assert browser.wait_for_not(query_string_fixture.is_state_now, 'showing nested responsive content')

    browser.click(XPath.input_labelled('Nested trigger field'))
    assert browser.wait_for(query_string_fixture.is_state_now, 'showing outer responsive content')
    assert browser.wait_for(query_string_fixture.is_state_now, 'showing nested responsive content')

    assert browser.wait_for(fixture.are_all_parts_enabled, browser)

    browser.click(XPath.input_labelled('Trigger field'))
    assert browser.wait_for_not(query_string_fixture.is_state_now, 'showing outer responsive content')
    assert browser.wait_for_not(query_string_fixture.is_state_now, 'showing nested responsive content')

    # Case: after having loaded the nested bits via ajax (and their Javascript - a second time)
    #  the nested case still works (necessary to test because loading JS a second time can cause bugs)
    browser.click(XPath.input_labelled('Trigger field'))
    assert browser.wait_for(query_string_fixture.is_state_now, 'showing outer responsive content')
    assert browser.wait_for_not(query_string_fixture.is_state_now, 'showing nested responsive content')

    browser.click(XPath.input_labelled('Nested trigger field'))
    assert browser.wait_for(query_string_fixture.is_state_now, 'showing outer responsive content')
    assert browser.wait_for(query_string_fixture.is_state_now, 'showing nested responsive content')

    assert browser.wait_for(fixture.are_all_parts_enabled, browser)


@with_fixtures(WebFixture, DisclosedInputFixture)
def test_clear_previously_given_user_input(web_fixture, disclosed_input_fixture):
    """If an input is refreshed as part of a Widget, and disappears, then reappears after a second refresh, its input is defaulted to the model value."""

    fixture = disclosed_input_fixture
    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MyForm.factory())

    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser

    user_input = 'valid.email@example.org'
    browser.open('/')

    browser.type(XPath.input_labelled('Email'), user_input)

    browser.click(XPath.input_labelled('Trigger field'))
    browser.wait_for_not(browser.is_visible, XPath.input_labelled('Email'))
    browser.click(XPath.input_labelled('Trigger field'))
    retained_value = browser.get_value(XPath.input_labelled('Email'))
    assert retained_value == ''
    assert retained_value != user_input


@with_fixtures(WebFixture, QueryStringFixture, ResponsiveDisclosureFixture)
def test_browser_back_after_state_changes_goes_to_previous_url(web_fixture, query_string_fixture, responsive_disclosure_fixture):
    """If a browser stores an URL in its history, which is the same as the one a user is currently on, 
       when you navigate history to get back to that URL, the page is rendered with newer state."""

    fixture = responsive_disclosure_fixture

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MainWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser

    class FormWithButton(fixture.MyForm):
        def __init__(self, view, an_object):
            super().__init__(view, an_object)

            self.define_event_handler(self.events.submit)
            self.add_child(ButtonInput(self, self.events.submit))

        events = ExposedNames()
        events.submit = lambda i: Event(label='Submit')

    fixture.MyForm = FormWithButton

    # create some state on the server side for this URL
    browser.open('/')
    assert browser.wait_for(query_string_fixture.is_state_now, 1)
    browser.select(XPath.select_labelled('Choice'), 'Two')
    assert browser.wait_for(query_string_fixture.is_state_now, 2)

    # submit the page, so we get redirected to the same URL, leaving itself in browser history (webkit only)
    browser.click(XPath.button_labelled('Submit'))
    assert browser.current_url.path == '/'

    # create different state on the same URL we were redirected to 
    browser.select(XPath.select_labelled('Choice'), 'Three')

    # when the user navigates back to the same url in history, the page retains the youngest state
    browser.refresh()
    assert browser.current_url.path == '/'
    assert browser.wait_for(query_string_fixture.is_state_now, 3)



@uses(query_string_fixture=QueryStringFixture)
class RecalculateWidgetFixture(Fixture):
    break_on_recalculate = False
    read_only = False
    def new_ModelObject(self):
        fixture = self
        class ModelObject(Base):
            __tablename__ = 'test_responsive_disclosure_recalculate'
            id = Column(Integer, primary_key=True)
            choice = Column(Integer)
            calculated_state = Column(Integer)

            def __init__(self):
                super().__init__()
                self.choice = 1
                self.recalculate()

            def recalculate(self):
                if fixture.break_on_recalculate and self.choice == 2:
                    raise DomainException(message='Breaking intentionally on recalculate')
                self.calculated_state = self.choice * 10

            def submit(self):
                raise DomainException(message='An exception happened on submit')

            events = ExposedNames()
            events.choice_changed = lambda i: Event(action=Action(i.recalculate))
            events.submit = lambda i: Event(action=Action(i.submit))

            fields = ExposedNames()
            fields.choice = lambda i: ChoiceField([Choice(1, IntegerField(label='One')),
                                                   Choice(2, IntegerField(label='Two')),
                                                   Choice(3, IntegerField(label='Three'))],
                                                  label='Choice')
            fields.calculated_state = lambda i: IntegerField(label='Calculated', writable=Allowed(not fixture.read_only))
        return ModelObject

    def new_model_object(self):
        return self.ModelObject()

    def new_MyForm(self):
        fixture = self
        class MyForm(Form):
            def __init__(self, view, an_object):
                super().__init__(view, 'myform')
                self.use_layout(FormLayout())
                self.an_object = an_object

                self.call_enable_refresh()

                if self.exception:
                    self.layout.add_alert_for_domain_exception(self.exception)
                self.change_trigger_input = TextInput(self, an_object.fields.choice, refresh_widget=self)
                self.layout.add_input(self.change_trigger_input)
                self.add_child(P(self.view, text='My state is now %s' % an_object.choice))
                fixture.add_to_form(self, an_object)
                self.define_event_handler(an_object.events.submit)
                self.add_child(ButtonInput(self, an_object.events.submit))

            def call_enable_refresh(self):
                self.enable_refresh(on_refresh=self.an_object.events.choice_changed)

            def recalculate(self):
                self.an_object.recalculate()


        return MyForm

    def new_MainWidgetWithPersistentModelObject(self):
        fixture = self
        class MainWidgetWithPersistentModelObject(Widget):
            def __init__(self, view):
                super().__init__(view)
                an_object = fixture.model_object
                self.add_child(fixture.MyForm(view, an_object))

        return MainWidgetWithPersistentModelObject


@uses(query_string_fixture=QueryStringFixture, recalculate_fixture=RecalculateWidgetFixture)
class RecalculatedWidgetScenarios(Fixture):

    @scenario
    def plain_widget(self):
        def add_to_form(form, model_object):
            form.add_child(P(form.view, text='My calculated state is now %s' % model_object.calculated_state))

        def check_widget_value(browser, value):
            browser.wait_for(self.query_string_fixture.is_state_labelled_now, 'My calculated state', value)

        self.recalculate_fixture.add_to_form = add_to_form
        self.check_widget_value = check_widget_value
        self.recalculate_fixture.read_only = True

    @scenario
    def writable_input(self):
        def add_to_form(form, model_object):
            text_input = form.layout.add_input(TextInput(form, model_object.fields.calculated_state))
            form.add_child(P(form.view, text='Status: %s' % text_input.get_input_status()))

        def check_widget_value(browser, value):
            browser.wait_for(browser.is_element_value, XPath.input_labelled('Calculated'), str(value))
            status_text = browser.get_text(XPath.paragraph().including_text('Status: '))
            assert 'invalidly_entered' not in status_text

        self.recalculate_fixture.add_to_form = add_to_form
        self.check_widget_value = check_widget_value
        self.recalculate_fixture.read_only = False

    @scenario
    def read_only_input(self):
        self.writable_input()
        self.recalculate_fixture.read_only = True


@with_fixtures(WebFixture, QueryStringFixture, SqlAlchemyFixture, RecalculateWidgetFixture, RecalculatedWidgetScenarios)
def test_recalculate_on_refresh(web_fixture, query_string_fixture, sql_alchemy_fixture, recalculate_fixture, scenario):
    """You can make a widget recalculate domain values upon refresh by adding an Event to enable_refresh()."""

    fixture = recalculate_fixture

    with sql_alchemy_fixture.persistent_test_classes(fixture.ModelObject):

        Session.add(fixture.model_object)

        wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MainWidgetWithPersistentModelObject.factory())
        web_fixture.reahl_server.set_app(wsgi_app)
        browser = web_fixture.driver_browser

        browser.open('/')
        assert browser.wait_for(query_string_fixture.is_state_now, 1)
        scenario.check_widget_value(browser, 10)
        browser.type(XPath.input_labelled('Choice'), '2')

        # Case: values are recalculated after ajax
        assert browser.wait_for(query_string_fixture.is_state_now, 2)
        scenario.check_widget_value(browser, 20)

        # Case: values stay recalculated after submit with exception
        browser.click(XPath.button_labelled('submit'))
        assert browser.is_element_present(XPath.paragraph().including_text('An exception happened on submit'))
        scenario.check_widget_value(browser, 20)


@with_fixtures(WebFixture, QueryStringFixture, SqlAlchemyFixture, RecalculateWidgetFixture)
def test_error_on_refresh_action(web_fixture, query_string_fixture, sql_alchemy_fixture, recalculate_fixture):
    """When a DomainException is raised during an Action executed on_refresh, the Widget should still render
       (albeit differently) despite the fact that there was an exception and all inputs causing the exception
       are retained.
    """

    fixture = recalculate_fixture
    class FormThatCatchesErrors(fixture.MyForm):
        can_calculate = True

        def call_enable_refresh(self):
            self.can_calculate = True
            try:
                super().call_enable_refresh()
            except DomainException as ex:
                self.layout.add_alert_for_domain_exception(ex)
                self.can_calculate = False

    def add_to_form(form, model_object):
        if form.can_calculate:
            form.add_child(P(form.view, text='My calculated state is now %s' % model_object.calculated_state))
        else:
            form.add_child(P(form.view, text='Invalid input, cannot calculate'))

    fixture.MyForm = FormThatCatchesErrors
    fixture.add_to_form = add_to_form

    with sql_alchemy_fixture.persistent_test_classes(fixture.ModelObject):

        Session.add(fixture.model_object)

        wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MainWidgetWithPersistentModelObject.factory())
        web_fixture.reahl_server.set_app(wsgi_app)
        browser = web_fixture.driver_browser

        browser.open('/')
        assert browser.wait_for(query_string_fixture.is_state_now, 1)
        fixture.break_on_recalculate = True

        # The Widget is rendered differently, and inputs retained.
        browser.type(XPath.input_labelled('Choice'), '2', wait_for_ajax=False)
        error_message = XPath.paragraph().including_text('Breaking intentionally on recalculate').inside_of(XPath.div().including_class('errors'))
        assert browser.is_element_present(error_message)
        modified_widget_message = XPath.paragraph().including_text('Invalid input, cannot calculate')
        assert browser.is_element_present(modified_widget_message)

        # After breakage, selecting a working value restored how the Widget renders itself
        browser.type(XPath.input_labelled('Choice'), '3', wait_for_ajax=False)
        assert browser.is_element_present(XPath.paragraph().including_text('My state is now 3'))
        assert browser.is_element_present(XPath.paragraph().including_text('My calculated state is now 30'))

        # Implementation corner case: Subsequent breakages work correctly again
        browser.type(XPath.input_labelled('Choice'), '2', wait_for_ajax=False)
        assert browser.is_element_present(XPath.paragraph().including_text('My state is now 2'))
        assert browser.is_element_present(XPath.paragraph().including_text('Invalid input, cannot calculate'))


class DebugScenarios(Fixture):
    @scenario
    def debug_on(self):
        self.expected_debug = True

    @scenario
    def debug_off(self):
        self.expected_debug = False


@with_fixtures(WebFixture, QueryStringFixture, SqlAlchemyFixture, RecalculateWidgetFixture, DebugScenarios)
def test_forgotten_error_on_refresh_action(web_fixture, query_string_fixture, sql_alchemy_fixture, recalculate_fixture, debug_scenario):
    """When an DomainException is raised during an Action executed on_refresh, and the Widget does not handle it,
       and error page is displayed.
    """

    fixture = recalculate_fixture
    def add_to_form(form, model_object):
        form.add_child(P(form.view, text='My calculated state is now %s' % model_object.calculated_state))

    fixture.add_to_form = add_to_form

    with sql_alchemy_fixture.persistent_test_classes(fixture.ModelObject):

        Session.add(fixture.model_object)

        wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.MainWidgetWithPersistentModelObject.factory())
        wsgi_app.config.reahlsystem.debug = debug_scenario.expected_debug
        web_fixture.reahl_server.set_app(wsgi_app)
        browser = web_fixture.driver_browser

        browser.open('/')
        assert browser.wait_for(query_string_fixture.is_state_now, 1)
        fixture.break_on_recalculate = True

        # Upon the ajax error, redirect to an error page
        with web_fixture.reahl_server.in_background(): # because we expect this call to throw an exception and we dont want to catch it here
            browser.type(XPath.input_labelled('Choice'), '2', wait_for_ajax=False)
            browser.wait_for_page_to_load()
        assert browser.wait_for_element_visible(XPath.heading(1).with_text('An error occurred:'))


@with_fixtures(WebFixture, QueryStringFixture, SqlAlchemyFixture)
def test_invalid_trigger_inputs(web_fixture, query_string_fixture, sql_alchemy_fixture):
    """Invalid values of trigger inputs are retained, while the calculated values are based on the last valid values in the domain."""

    fixture = scenario

    class ModelObject(Base):
        __tablename__ = 'test_responsive_disclosure_recalculate_invalids'
        id = Column(Integer, primary_key=True)
        choice = Column(Integer, default=1)
        choice2 = Column(Integer, default=4)
        choice3 = Column(Integer, default=9)
        calculated_state = Column(Integer, default=0)

        def recalculate(self):
            self.calculated_state = self.choice + self.choice2

        def submit(self):
            raise DomainException(message='An exception happened on submit')

        events = ExposedNames()
        events.choice_changed = lambda i: Event(action=Action(i.recalculate))
        events.submit = lambda i: Event(action=Action(i.submit))

        fields = ExposedNames()
        fields.choice = lambda i: ChoiceField([Choice(1, IntegerField(label='One')),
                                               Choice(2, IntegerField(label='Two')),
                                               Choice(3, IntegerField(label='Three'))],
                                              label='Choice')
        fields.choice2 = lambda i: ChoiceField([Choice(4, IntegerField(label='Four')),
                                                Choice(5, IntegerField(label='Five')),
                                                Choice(6, IntegerField(label='Six'))],
                                               label='Choice2')
        fields.calculated_state = lambda i: IntegerField(label='Calculated', writable=Allowed(False))

    class MyForm(Form):
        def __init__(self, view, an_object):
            super().__init__(view, 'myform')
            self.use_layout(FormLayout())
            self.an_object = an_object
            self.enable_refresh(on_refresh=an_object.events.choice_changed)
            if self.exception:
                self.layout.add_alert_for_domain_exception(self.exception)
            self.change_trigger_input = TextInput(self, an_object.fields.choice, refresh_widget=self)
            self.layout.add_input(self.change_trigger_input)
            self.add_child(P(self.view, text='My choice state is now %s' % an_object.choice))
            self.change2_trigger_input = TextInput(self, an_object.fields.choice2, refresh_widget=self)
            self.layout.add_input(self.change2_trigger_input)
            self.add_child(P(self.view, text='My choice2 state is now %s' % an_object.choice2))
            self.add_child(P(self.view, text='My calculated state is now %s' % an_object.calculated_state))
            self.define_event_handler(an_object.events.submit)
            self.add_child(ButtonInput(self, an_object.events.submit))

    class MainWidgetWithPersistentModelObject(Widget):
        def __init__(self, view):
            super().__init__(view)
            an_object = fixture.model_object
            self.add_child(MyForm(view, an_object))


    with sql_alchemy_fixture.persistent_test_classes(ModelObject):

        fixture.model_object = ModelObject()
        Session.add(fixture.model_object)

        wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=MainWidgetWithPersistentModelObject.factory())
        web_fixture.reahl_server.set_app(wsgi_app)
        browser = web_fixture.driver_browser

        browser.open('/')
        assert browser.wait_for(query_string_fixture.is_state_labelled_now, 'My choice state', 1)
        assert browser.wait_for(query_string_fixture.is_state_labelled_now, 'My choice2 state', 4)
        assert browser.wait_for(query_string_fixture.is_state_labelled_now, 'My calculated state', '5')

        # Case: Entering an invalid value does not trigger a refresh of the input doing the triggering
        with web_fixture.driver_browser.no_load_expected_for('input[name="myform-choice"]'):
            browser.type(XPath.input_labelled('Choice'), 'invalid value')

        # Case: Entering an valid value in a different trigger, triggers a refresh, but last valid value of choice is used
        browser.type(XPath.input_labelled('Choice2'), '5')
        assert browser.wait_for(query_string_fixture.is_state_labelled_now, 'My calculated state', '6')

        #       But, the invalid input is retained
        assert browser.is_element_value(XPath.input_labelled('Choice'), 'invalid value')


@with_fixtures(WebFixture, QueryStringFixture, SqlAlchemyFixture)
def test_invalid_non_trigger_input_corner_case(web_fixture, query_string_fixture, sql_alchemy_fixture):
    """If an invalid value was submitted via ajax for a non-trigger input, and a valid value is submitted for is with a 
       form, and a DomainException happens... then the non-trigger input must retain its new, valid value."""

    fixture = scenario

    class ModelObject(Base):
        __tablename__ = 'test_responsive_disclosure_recalculate_invalids'
        id = Column(Integer, primary_key=True)
        choice = Column(Integer, default=1)
        choice3 = Column(Integer, default=9)
        calculated_state = Column(Integer, default=0)

        def recalculate(self):
            self.calculated_state = self.choice * 10

        def submit(self):
            raise DomainException(message='An exception happened on submit')

        events = ExposedNames()
        events.choice_changed = lambda i: Event(action=Action(i.recalculate))
        events.submit = lambda i: Event(action=Action(i.submit))

        fields = ExposedNames()
        fields.choice = lambda i: ChoiceField([Choice(1, IntegerField(label='One')),
                                               Choice(2, IntegerField(label='Two')),
                                               Choice(3, IntegerField(label='Three'))],
                                              label='Choice')
        fields.choice3 = lambda i: ChoiceField([Choice(7, IntegerField(label='Seven')),
                                                Choice(8, IntegerField(label='Eight')),
                                                Choice(9, IntegerField(label='Nine'))],
                                               label='Choice3')
        fields.calculated_state = lambda i: IntegerField(label='Calculated', writable=Allowed(False))

    class MyForm(Form):
        def __init__(self, view, an_object):
            super().__init__(view, 'myform')
            self.an_object = an_object
            self.use_layout(FormLayout())
            self.enable_refresh(on_refresh=an_object.events.choice_changed)
            if self.exception:
                self.add_child(P(self.view, text=str(self.exception)))
            self.change_trigger_input = self.layout.add_input(TextInput(self, an_object.fields.choice, refresh_widget=self))
            self.add_child(P(self.view, text='My choice state is now %s' % an_object.choice))
            self.change3_non_trigger_input = self.layout.add_input(TextInput(self, an_object.fields.choice3))
            self.add_child(P(self.view, text='My calculated state is now %s' % an_object.calculated_state))
            self.define_event_handler(an_object.events.submit)
            self.add_child(ButtonInput(self, an_object.events.submit))

    class MainWidgetWithPersistentModelObject(Widget):
        def __init__(self, view):
            super().__init__(view)
            an_object = fixture.model_object
            self.add_child(MyForm(view, an_object))


    with sql_alchemy_fixture.persistent_test_classes(ModelObject):

        fixture.model_object = ModelObject()
        Session.add(fixture.model_object)

        wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=MainWidgetWithPersistentModelObject.factory())
        web_fixture.reahl_server.set_app(wsgi_app)
        browser = web_fixture.driver_browser

        browser.open('/')
        assert browser.wait_for(query_string_fixture.is_state_labelled_now, 'My choice state', 1)
        assert browser.wait_for(query_string_fixture.is_state_labelled_now, 'My calculated state', '10')

        browser.type(XPath.input_labelled('Choice3'), 'other invalid input')
        browser.type(XPath.input_labelled('Choice'), '2')
        browser.press_tab()
        assert browser.wait_for(query_string_fixture.is_state_labelled_now, 'My calculated state', '20')

        assert browser.is_element_value(XPath.input_labelled('Choice3'), 'other invalid input')
        browser.type(XPath.input_labelled('Choice3'), '8')
        browser.click(XPath.button_labelled('submit'))
        assert browser.is_element_present(XPath.paragraph().including_text('An exception happened on submit'))
        assert browser.is_element_value(XPath.input_labelled('Choice3'), '8')


# TODO: What about when you have multiple Forms on a page, with an input on both Forms that uses a Field bound to the same name on the same domain object?
#       Upon Ajax, all form inputs are submitted, not just one. It may be that their processing is ordered such that the one that never refreshes always happens last.
#       Because it happens last, it always overwrites the value to the old non-refreshed value and nothing ever changes?
#       One idea: first accept_input for all inputs outside of the widget being refreshed. Then the inputs in that Widget. Then the actual input which triggered the change. ?



# Copyright 2013-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from webob import Request

from reahl.stubble import stubclass
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import Browser

from reahl.web.fw import UserInterface, Widget, FactoryDict, UserInterfaceFactory, RegexPath, UrlBoundView
from reahl.web.ui import HTML5Page, P, A, Div, Slot

from reahl.web_dev.fixtures import WebFixture, BasicPageLayout


@with_fixtures(WebFixture)
def test_basic_ui(web_fixture):
    """A UserInterface is a chunk of web app that can be grafted onto the URL hierarchy of any app.

       A UserInterface has its own views. Its Views are relative to the UserInterface itself.
    """
    class UIWithTwoViews(UserInterface):
        def assemble(self):
            self.define_view('/', title='UserInterface root view')
            self.define_view('/other', title='UserInterface other view')

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page)
            self.define_user_interface('/a_ui',  UIWithTwoViews,  {}, name='myui')

    fixture = web_fixture
    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/a_ui/')
    assert browser.title == 'UserInterface root view' 

    browser.open('/a_ui/other')
    assert browser.title == 'UserInterface other view' 


@with_fixtures(WebFixture)
def test_ui_slots_map_to_window(web_fixture):
    """The UserInterface uses its own names for Slots. When attaching a UserInterface, you have to specify 
        which of the UserInterface's Slots plug into which of the page's Slots.
    """
    class UIWithSlots(UserInterface):
        def assemble(self):
            root = self.define_view('/', title='UserInterface root view')
            root.set_slot('text', P.factory(text='in user_interface slot named text'))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            self.define_user_interface('/a_ui',  UIWithSlots,  {'text': 'main'}, name='myui')

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/a_ui/')
    assert browser.title == 'UserInterface root view' 

    # The widget in the UserInterface's slot named 'text' end up in the HTML5Page slot called main
    [p] = browser.lxml_html.xpath('//div[contains(@class,"column-main")]/p')
    assert p.text == 'in user_interface slot named text' 


@with_fixtures(WebFixture)
def test_ui_redirect(web_fixture):
    """When opening an URL without trailing slash that maps to where a UserInterface is attached,
       the browser is redirected to the UserInterface '/' View."""

    class UIWithRootView(UserInterface):
        def assemble(self):
            self.define_view('/', title='UserInterface root view')

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page)
            self.define_user_interface('/a_ui',  UIWithRootView,  {}, name='myui')

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/a_ui')
    assert browser.title == 'UserInterface root view' 
    assert browser.current_url.path == '/a_ui/' 


@with_fixtures(WebFixture)
def test_ui_arguments(web_fixture):
    """UserInterfaces can take exta args and kwargs."""

    class UIWithArguments(UserInterface):
        def assemble(self, kwarg=None):
            self.kwarg = kwarg
            text = self.kwarg
            root = self.define_view('/', title='A view')
            root.set_slot('text', P.factory(text=text))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            self.define_user_interface('/a_ui', UIWithArguments, {'text': 'main'},
                            name='myui', kwarg='the kwarg')

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/a_ui/')
    [p] = browser.lxml_html.xpath('//p')
    assert p.text == 'the kwarg' 


@with_fixtures(WebFixture)
def test_bookmarks(web_fixture):
    """Bookmarks are pointers to Views. You need them, because Views are relative to a UserInterface and
       a Bookmark can, at run time, turn these into absolute URLs. Bookmarks also contain metadata,
       such as the title of the View it points to.
    """
    fixture = web_fixture
    user_interface = UserInterface(None, '/a_ui', {}, False, 'test_ui')
    view = UrlBoundView(user_interface, '/aview', 'A View title')
    bookmark = view.as_bookmark()

    # What the bookmark knows
    assert bookmark.href.path == '/a_ui/aview' 
    assert bookmark.description == 'A View title' 
    assert bookmark.base_path == '/a_ui' 
    assert bookmark.relative_path == '/aview' 

    # How you would use a bookmark in other views (possibly in other UserInterfaces)
    a = A.from_bookmark(fixture.view, bookmark)
    assert str(a.href) == str(bookmark.href) 


@with_fixtures(WebFixture)
def test_bookmarks_overrides(web_fixture):
    """Various bits of information can be overridden from the defaults when creating a bookmark from a View.
    """
    fixture = web_fixture

    user_interface = UserInterface(None, '/a_ui', {}, False, 'test_ui')
    view = UrlBoundView(user_interface, '/aview', 'A View title')
    bookmark = view.as_bookmark(description='different description',
                                query_arguments={'arg1': 'val1'},
                                locale='af')

    # What the bookmark knows
    assert bookmark.description == 'different description' 
    assert bookmark.query_arguments == {'arg1': 'val1'} 
    assert bookmark.locale == 'af' 


@with_fixtures(WebFixture)
def test_bookmarks_from_other_sources(web_fixture):
    """Bookmarks can also be made from ViewFactories, UserInterfaces or UserInterfaceFactories. 
    """
    fixture = web_fixture

    class UIWithRelativeView(UserInterface):
        def assemble(self):
            view_factory = self.define_view('/aview', title='A View title')

            # How you could get a bookmark from a UserInterface or ViewFactory
            fixture.bookmark_from_view_factory = view_factory.as_bookmark(self)
            fixture.bookmark_from_ui = self.get_bookmark(relative_path='/aview')

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page)
            fixture.ui_factory = self.define_user_interface('/a_ui',  UIWithRelativeView,  {}, name='myui')

            # How you could get a bookmark from a UserInterfaceFactory
            fixture.bookmark_from_ui_factory = fixture.ui_factory.get_bookmark(relative_path='/aview')

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    Browser(wsgi_app).open('/a_ui/aview') # To execute the above once

    for bookmark in [fixture.bookmark_from_view_factory, 
                     fixture.bookmark_from_ui, 
                     fixture.bookmark_from_ui_factory]:
        # What the bookmark knows
        assert bookmark.href.path == '/a_ui/aview' 
        assert bookmark.description == 'A View title' 
        assert bookmark.base_path == '/a_ui' 
        assert bookmark.relative_path == '/aview' 


@with_fixtures(WebFixture)
def test_the_lifecycle_of_a_ui(web_fixture):
    """This test illustrates the steps a UserInterface goes through from being specified, to
       being used. It tests a couple of lower-level implementation issues (see comments)."""

    def current_view_is_plugged_in(page):
        return page.slot_contents['main_slot'].__class__ is Div

    @stubclass(UserInterface)
    class UserInterfaceStub(UserInterface):
        assembled = False
        def assemble(self, **ui_arguments):
            self.controller_at_assemble_time = self.controller
            root = self.define_view('/some/path', title='A view')
            root.set_slot('slotA', Div.factory())
            self.assembled = True

    fixture = web_fixture

    # Phase1: specifying a user_interface and assembleing it to a site (with kwargs)
    parent_ui = None
#        parent_ui = EmptyStub(base_path='/')
    slot_map = {'slotA': 'main_slot'}
    ui_factory = UserInterfaceFactory(parent_ui, RegexPath('/', '/', {}), slot_map, UserInterfaceStub, 'test_ui')


    # Phase2: creating a user_interface
    request = Request.blank('/some/path')
    fixture.context.request = request
    user_interface = ui_factory.create('/some/path', for_bookmark=False)

    # - Assembly happened correctly
    assert user_interface.parent_ui is parent_ui 
    assert user_interface.slot_map is slot_map 
    assert user_interface.name is 'test_ui' 
    assert user_interface.relative_base_path == '/' 
    assert user_interface.controller_at_assemble_time is not None
    assert user_interface.controller is not None 
    assert user_interface.assembled 

    # - Create for_bookmark empties the relative_path
    user_interface = ui_factory.create('/some/path', for_bookmark=True)
    assert user_interface.relative_path == '' 

    # - The relative_path is reset if not done for_bookmark
    #   This is done in case a for_bookmark call just
    #   happened to be done for the same UserInterface in the same request
    #   before the "real" caal is done
    user_interface = ui_factory.create('/some/path', for_bookmark=False)
    assert user_interface.relative_path == '/some/path' 

    # Phase5: create the page and plug the view into it
    page = Widget.factory().create(user_interface.current_view)
    page.add_child(Slot(user_interface.current_view, 'main_slot'))

    page.plug_in(user_interface.current_view)
    assert current_view_is_plugged_in(page) 
    assert isinstance(user_interface.sub_resources, FactoryDict) 



# Copyright 2013-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.




from reahl.tofu import expected
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import Browser

from reahl.component.modelinterface import Field, RequiredConstraint
from reahl.component.exceptions import ProgrammerError
from reahl.web.fw import UrlBoundView, UserInterface
from reahl.web.ui import HTML5Page

from reahl.web_dev.fixtures import WebFixture


@with_fixtures(WebFixture)
def test_missing_variable_in_regex(web_fixture):
    """If a variable is missing from the regex, an appropriate error is raised."""

    class ParameterisedView(UrlBoundView):
        def assemble(self, some_key=None):
            self.title = 'View for: %s' % some_key

    class UIWithParameterisedViews(UserInterface):
        def assemble(self):
            self.define_regex_view('/(?P<incorrect_name_for_key>.*)', '/${key}', view_class=ParameterisedView, some_key=Field(required=True))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page)
            self.define_user_interface('/a_ui',  UIWithParameterisedViews,  {}, name='test_ui')

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    with expected(ProgrammerError, test='The arguments contained in URL.*'):
        browser.open('/a_ui/test1/')


@with_fixtures(WebFixture)
def test_missing_variable_in_ui_regex(web_fixture):

    class RegexUserInterface(UserInterface):
        def assemble(self, ui_key=None):
            self.name = 'user_interface-%s' % ui_key

    class UIWithParameterisedUserInterfaces(UserInterface):
        def assemble(self):
            self.define_regex_user_interface('/(?P<xxx>[^/]*)', 'N/A', RegexUserInterface,
                                             {},
                                             ui_key=Field(required=True))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page)
            self.define_user_interface('/a_ui',  UIWithParameterisedUserInterfaces,  {}, name='test_ui')


    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)

    browser = Browser(wsgi_app)

    with expected(RequiredConstraint):
        browser.open('/a_ui/test1/')



# Copyright 2014-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from selenium.webdriver.common.by import By

from reahl.stubble import EmptyStub
from reahl.tofu import Fixture, scenario, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import XPath, WidgetTester
from reahl.web.fw import Widget
from reahl.web.ui import StaticColumn, DynamicColumn, Table, Thead, Span, Div, P, TextNode

from reahl.component.modelinterface import Field, BooleanField

from reahl.web_dev.fixtures import WebFixture
from reahl.dev.fixtures import ReahlSystemFixture


class DataItem:
    def __init__(self, row, alpha):
        self.row = row
        self.alpha = alpha


@uses(web_fixture=WebFixture)
class TableFixture(Fixture):

    def new_data(self):
        return [DataItem(1, 'T'), DataItem(2, 'H'), DataItem(3, 'E')]

    def table_caption_is(self, expected):
        return  self.web_fixture.driver_browser.find_element(XPath.caption().with_text(expected))

    def table_summary_is(self, summary):
        return  self.web_fixture.driver_browser.find_element(XPath.table_with_summary(summary))

    def table_column_name_is(self, column_number, expected):
        return self.web_fixture.driver_browser.web_driver.find_element(By.XPATH, '((//table/thead/tr/th)[%s]/span)[1]' % column_number).text == expected

    def get_table_row(self, row_number):
        row_data = []
        for column_number in list(range(1, self.table_number_columns()+1)):
            row_data.append(self.web_fixture.driver_browser.web_driver.find_element(By.XPATH, '((//table/tbody/tr)[%s]/td)[%s]' % (row_number, column_number)).text)
        return row_data

    def table_number_columns(self):
        return len(self.web_fixture.driver_browser.web_driver.find_elements(By.XPATH, '//table/thead/tr/th'))

    def table_number_rows(self):
        return len(self.web_fixture.driver_browser.web_driver.find_elements(By.XPATH, '//table/tbody/tr'))


@with_fixtures(WebFixture, TableFixture)
def test_table_basics(web_fixture, table_fixture):
    """A Table populated .with_data() displays a list of items as defined by a list of Columns"""

    class MainWidget(Div):
        def __init__(self, view):
            super().__init__(view)
            table = Table(view, caption_text='All my friends', summary='Summary for screen reader')
            table.with_data(
                            [StaticColumn(Field(label='Row Number'), 'row'),
                             StaticColumn(Field(label='Alpha'), 'alpha')],
                             table_fixture.data
                            )
            self.add_child(table)

    fixture = table_fixture

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=MainWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    web_fixture.driver_browser.open('/')

    # The table has a caption and summary
    assert fixture.table_caption_is('All my friends')
    assert fixture.table_summary_is('Summary for screen reader')

    # Column headings are derived from given Column Fields
    assert fixture.table_column_name_is(1, 'Row Number')
    assert fixture.table_column_name_is(2, 'Alpha')

    # A string representation of the value of each Field of a given data item is shown in the appropriate cell
    assert fixture.table_number_rows() == 3

    assert fixture.get_table_row(1) == ['1', 'T']
    assert fixture.get_table_row(2) == ['2', 'H']
    assert fixture.get_table_row(3) == ['3', 'E']


@uses(reahl_system_fixture=ReahlSystemFixture)
class Scenarios(Fixture):
    @scenario
    def static_column(self):
        self.total_column = StaticColumn(Field(label='Alpha'), 'alpha', footer_label='Total')
        self.expected_total = 'Total'

    @scenario
    def dynamic_column(self):
        self.total_column = DynamicColumn('Heading', lambda view, item: Widget(view), 
                                          make_footer_widget=lambda view, item: TextNode(view, str(item.total)))
        self.expected_total = '123'


@with_fixtures(WebFixture, TableFixture, Scenarios)
def test_table_totals(web_fixture, table_fixture, scenario):
    """You can pass footer_items to with_data to add a footer with a row for each footer_item.
    Each column with footer content defined will have that content rendered in its footer row,
    columns inbetween will be collapsed with a colspan to ensure alignment.
    """

    class MainWidget(Div):
        def __init__(self, view):
            super().__init__(view)
            table = Table(view)
            table.with_data(
                            [StaticColumn(Field(label='Another Row Number'), 'row_another'),
                             StaticColumn(Field(label='Row Number'), 'row'),
                             scenario.total_column],
                             table_fixture.data,
                             footer_items=[EmptyStub(total=123)]
                            )
            self.add_child(table)

    fixture = table_fixture

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=MainWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    filler_cell = XPath.table_cell()[1].inside_of(XPath.table_footer())
    assert browser.get_attribute(filler_cell, 'colspan') == '2'
    total_cell = XPath.table_cell()[2].inside_of(XPath.table_footer())
    assert browser.get_text(total_cell) == scenario.expected_total
    assert not browser.get_attribute(total_cell, 'colspan') 


@uses(web_fixture=WebFixture)
class ColumnScenarios(Fixture):
    sort_key = EmptyStub()
    heading = 'A heading'
    row_item = EmptyStub(some_attribute=True)

    @scenario
    def static_column(self):
        """StaticColumn represents an attribute of the item in a row, using a Field to translate the value of that attribute into a string and to specify a column header."""
        self.column = StaticColumn(BooleanField(label=self.heading), 'some_attribute', sort_key=self.sort_key)
        self.expected_cell_html = 'on' # as translated by BooleanField
        self.expected_heading_html = '<span>A heading</span>'

    @scenario
    def dynamic_column(self):
        """DynamicColumn uses a function to create a Widget on the fly for each cell and uses the specified heading."""
        def make_span(view, data_item):
            return Span(view, text='Answer: %s' % (data_item.some_attribute))

        self.column = DynamicColumn(self.heading, make_span, sort_key=self.sort_key)
        self.expected_cell_html = '<span>Answer: True</span>' # raw attribute used
        self.expected_heading_html = '<span>A heading</span>'

    @scenario
    def dynamic_column_static_heading(self):
        """A DynamicColumn can also use a function to create its heading on the fly."""
        def make_span(view, data_item):
            return Span(view, text='Answer: %s' % (data_item.some_attribute))

        def make_heading(view):
            return P(view, text=self.heading)

        self.column = DynamicColumn(make_heading, make_span, sort_key=self.sort_key)
        self.expected_cell_html = '<span>Answer: True</span>' # raw attribute used
        self.expected_heading_html = '<p>A heading</p>'


@with_fixtures(WebFixture, ColumnScenarios)
def test_different_kinds_of_columns(web_fixture, column_scenarios):
    """There are different kinds of Columns, allowing different levels of flexibility for defining a Table"""

    fixture = column_scenarios

    assert fixture.column.sort_key is fixture.sort_key

    # The heading
    widget_for_heading = fixture.column.heading_as_widget(web_fixture.view)
    actual = WidgetTester(widget_for_heading).render_html()

    assert actual == fixture.expected_heading_html

    # A cell
    widget_for_cell = fixture.column.as_widget(web_fixture.view, fixture.row_item)
    actual = WidgetTester(widget_for_cell).render_html()

    assert actual == fixture.expected_cell_html


@with_fixtures(WebFixture)
def test_table_thead(web_fixture):
    """Table can find its Thead element"""


    table = Table(web_fixture.view)
    thead = table.add_child(Thead(web_fixture.view))

    assert table.thead is thead



# Copyright 2013-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



import os.path

from reahl.tofu import temp_file_with
from reahl.tofu.pytestsupport import with_fixtures

from reahl.component.modelinterface import FileField, ExposedNames, Event, UploadedFile, ValidationConstraint
from reahl.web.ui import SimpleFileInput, Form, ButtonInput
from reahl.browsertools.browsertools import XPath

from reahl.web_dev.fixtures import WebFixture


class FailingConstraint(ValidationConstraint):
    fail = True

    def validate_input(self, unparsed_input):
        if self.fail:
            raise self


@with_fixtures(WebFixture)
def test_form_encoding(web_fixture):
    """The enctype of a Form changes to multipart/form-data if it contains an input for a file."""
    fixture = web_fixture

    class DomainObject:
        fields = ExposedNames()
        fields.file = lambda i: FileField(allow_multiple=False, label='Attached files')

    domain_object = DomainObject()

    form = Form(fixture.view, 'testform')
    assert 'enctype' not in form.attributes.v

    form.add_child(SimpleFileInput(form, domain_object.fields.file))
    assert form.attributes.v['enctype'] == 'multipart/form-data'


@with_fixtures(WebFixture)
def test_simple_file_input(web_fixture):
    """A SimpleFileInput is a Widget with which a user can choose one or more files.
       The SimpleFileInput transforms the chosen files to UploadedFile objects, and passes these
       to its associated FileField upon a Form submit."""


    expected_content = b'some content'
    file_to_upload = temp_file_with(expected_content, mode='w+b')
    class DomainObject:
        def __init__(self):
           self.file = None

        fields = ExposedNames()
        fields.file = lambda i: FileField(allow_multiple=False, label='Attached files')

        events = ExposedNames()
        events.upload = lambda i: Event(label='Upload')

    domain_object = DomainObject()

    class FileUploadForm(Form):
        def __init__(self, view):
            super().__init__(view, 'test')
            self.add_child(SimpleFileInput(self, domain_object.fields.file))
            self.define_event_handler(domain_object.events.upload)
            self.add_child(ButtonInput(self, domain_object.events.upload))

    wsgi_app = web_fixture.new_wsgi_app(child_factory=FileUploadForm.factory(), enable_js=False)
    web_fixture.reahl_server.set_app(wsgi_app)

    browser = web_fixture.driver_browser
    browser.open('/')

    browser.type(XPath.input_of_type('file'), file_to_upload.name)
    browser.click(XPath.button_labelled('Upload'))
    assert isinstance(domain_object.file, UploadedFile)
    assert domain_object.file.filename == os.path.basename(file_to_upload.name)
    with domain_object.file.open() as opened_file:
        read_contents = opened_file.read()
    assert read_contents == expected_content


@with_fixtures(WebFixture)
def test_simple_file_input_exceptions(web_fixture):
    """Usually, when a DomainException happens during a form submit Inputs save the input they received so that
       such input can be pre-populated on the screen rendered by a subsequent GET for a user to correct
       possible mistakes and resubmit.

       In the case of a SimpleFileInput however, the UploadedFile objects that serve as user input
       to a SimpleFileInput are discarded when a DomainException happens.  This is because it is not
       possible to prepopulate its value using HTML or JS as this would be a security risk.
    """


    file_to_upload = temp_file_with('some content')
    failing_constraint = FailingConstraint('I am breaking')

    class DomainObject:
        def __init__(self):
            self.file = None

        fields = ExposedNames()
        def make_field(self):
            field = FileField(allow_multiple=False, label='Attached files')
            # FailingConstraint is declared in module level scope for it to be pickleable
            field.add_validation_constraint(failing_constraint)
            return field
        fields.file = make_field
        events = ExposedNames()
        events.upload = lambda i: Event(label='Upload')

    domain_object = DomainObject()

    class FileUploadForm(Form):
        def __init__(self, view):
            super().__init__(view, 'test')
            file_input = self.add_child(SimpleFileInput(self, domain_object.fields.file))
            if file_input.validation_error:
                self.add_child(self.create_error_label(file_input))
            self.define_event_handler(domain_object.events.upload)
            self.add_child(ButtonInput(self, domain_object.events.upload))

    wsgi_app = web_fixture.new_wsgi_app(child_factory=FileUploadForm.factory(), enable_js=False)
    web_fixture.reahl_server.set_app(wsgi_app)

    browser = web_fixture.driver_browser
    browser.open('/')
    
    browser.type(XPath.input_of_type('file'), file_to_upload.name)
    browser.click(XPath.button_labelled('Upload'))
    assert browser.is_element_present('//label[text()="I am breaking"]')

    # Message is cleared on second attempt
    failing_constraint.fail = False
    browser.type(XPath.input_of_type('file'), file_to_upload.name)
    browser.click(XPath.button_labelled('Upload'))
    assert not browser.is_element_present('//label[text()="I am breaking"]')



# Copyright 2016-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



from reahl.tofu import expected
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import WidgetTester
from reahl.component.exceptions import ProgrammerError

from reahl.web.holder.holder import PlaceholderImage, PredefinedTheme, CustomTheme

from reahl.web_dev.fixtures import WebFixture


@with_fixtures(WebFixture)
def test_placeholder_basics(web_fixture):
    """
       hint: Ensure the Holder(Library) is added to the  web.frontend_libraries config setting in the file:web.config.py
    """
    placeholder = PlaceholderImage(web_fixture.view, 20, 30)

    tester = WidgetTester(placeholder)

    expected_html = '<img data-src="holder.js/20x30">'
    actual = tester.render_html()
    assert actual == expected_html


@with_fixtures(WebFixture)
def test_placeholder_with_text(web_fixture):

    placeholder = PlaceholderImage(web_fixture.view, 20, 30, text='My banner')

    expected_value = 'holder.js/20x30?text=My banner'
    actual_value = placeholder.get_attribute('data-src')
    assert actual_value == expected_value


@with_fixtures(WebFixture)
def test_placeholder_with_predefine_theme(web_fixture):

    my_theme = PredefinedTheme('lava')
    placeholder = PlaceholderImage(web_fixture.view, 20, 30, theme=my_theme)

    expected_value = 'holder.js/20x30?theme=lava'
    actual_value = placeholder.get_attribute('data-src')
    assert actual_value == expected_value


@with_fixtures(WebFixture)
def test_text_and_theme_options_are_encoded(web_fixture):

    my_theme = CustomTheme(bg='white', fg='red')
    placeholder = PlaceholderImage(web_fixture.view, 20, 30, text='My sê goed', theme=my_theme)

    expected_value = 'holder.js/20x30?bg=white&fg=red&text=My sê goed'
    actual_value = placeholder.get_attribute('data-src')
    assert actual_value == expected_value


def test_custom_theme_options_become_text():

    my_theme = CustomTheme(bg='yellow', fg='blue',
                           text_size=12, text_font='arial', text_align='left',
                           outline='yes', line_wrap=0.5)

    options_dict = my_theme

    assert options_dict.get('bg') == 'yellow'
    assert options_dict.get('fg') == 'blue'
    assert options_dict.get('size') == '12'
    assert options_dict.get('font') == 'arial'
    assert options_dict.get('align') == 'left'
    assert options_dict.get('outline') == 'yes'
    assert options_dict.get('lineWrap') == '0.5'


def test_custom_theme_text_align_valid_values():

    #valid options
    CustomTheme(text_align='left')
    CustomTheme(text_align='right')
    CustomTheme(text_align=None)

    #invalid
    with expected(ProgrammerError):
        invalid_value = 'somewhere'
        CustomTheme(text_align=invalid_value)



# Copyright 2021, 2022, 2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Support for including `plotly.py <https://github.com/plotly/plotly.py/>`_ figures on a web page.
"""


import json

from reahl.web.ui import HTMLWidget, Div, LiteralHTML
from reahl.web.libraries import PlotlyJS


class RefreshedJS(Div):
    def __init__(self, chart):
        super().__init__(chart.view, css_id='%s-data-container' % chart.css_id)
        self.chart = chart

    def get_js(self, context=None):
        options = { 'chart_id': self.chart.css_id,
                    'json': json.loads(self.chart.figure.to_json())}
        return super().get_js(context=context) + ['$(%s).plotlychartdata(%s);' % (self.jquery_selector,
                                                                                  json.dumps(options))]

class ChartContents(Div):
    def __init__(self, chart):
        super().__init__(chart.view, css_id='%s-data' % chart.css_id)
        self.div = self.add_child(RefreshedJS(chart))
        self.enable_refresh()


class Chart(HTMLWidget):
    """A Widget displays a plotly Figure.

    :param view: (See :class:`reahl.web.fw.Widget`)
    :param figure: A fully constructed plotly Figure.
    :param css_id: A unique id for this Chart
    :keyword read_check: (See :class:`reahl.web.fw.Widget`)
    :keyword write_check: (See :class:`reahl.web.fw.Widget`)

    """
    def __init__(self, view, figure, css_id, read_check=None, write_check=None):
        super().__init__(view, read_check=read_check, write_check=write_check)
        self.figure = figure
        containing_div = self.add_child(Div(view, css_id=css_id))
        containing_div.append_class('reahl-plotlychart')
        self.set_html_representation(containing_div)
        self.enable_refresh()
        containing_div.add_child(LiteralHTML(view, PlotlyJS.get_instance().inline_material()))
        self.contents = containing_div.add_child(ChartContents(self)) #: The contents of the graph. You can use this as
                                                                      #: refresh_widget on an Input to refresh only the
                                                                      #: contents of the Chart, and not the entire Chart.

    def get_js(self, context=None):
        options = {'json': json.loads(self.figure.to_json())}
        return super().get_js(context=context) + ['$(%s).plotlychart(%s);' % (self.html_representation.jquery_selector,
                                                                              json.dumps(options))]



# Copyright 2015-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. versionadded:: 3.2

Carousel presents a slideshow of Widgets with captions on each slide.


"""

from reahl.web.fw import Widget, Url
from reahl.web.ui import HTMLAttributeValueOption, HTMLElement, Img
from reahl.web.bootstrap.ui import Div, A, Span, Li, Ol, TextNode
from reahl.component.i18n import Catalogue

_ = Catalogue('reahl-web')


class Slide(Div):
    def __init__(self, view, widget, caption_widget, index):
        super().__init__(view)
        self.index = index
        self.append_class('carousel-item')
        self.add_child(widget)
        if self.is_active:
            self.append_class('active')

        if isinstance(widget, Img):
            widget.append_class('d-block')
            widget.append_class('w-100')

        if caption_widget:
            self.add_child(self.create_caption(caption_widget))

    @property
    def is_active(self):
        return self.index == 0

    def create_caption(self, caption_widget):
        div = Div(self.view)
        div.append_class('carousel-caption')
        div.add_child(caption_widget)
        return div


class Carousel(Widget):
    """
    A Widget that appears as having multiple panels of content of which
    only one panel is visible at a time. The visible panel is
    exchanged for another panel of content when prompted or
    automatically, at regular intervals, by visually sliding the new
    panel into position thus creating a slideshow effect.

    :param view: (See :class:`~reahl.web.fw.Widget`)
    :param css_id: (See :class:`~reahl.web.ui.HTMLElement`)
    :keyword show_indicators: If True (the default), includes indicators showing the number of slides and which one is visible.
    :keyword interval: The number (an int) of milliseconds to delay between cycling content. If not given (the default), the Carousel will not cycle automatically.
    :keyword pause: If set to 'false', hovering over the carousel will not pause it. Another option, 'hover' is supported.
    :keyword wrap: If True, the Carousel cycles contiuously, else it stops at the last slide.
    :keyword keyboard: If True, the Carousel reacts to keyboard events.
    :keyword min_height: If given, force slides to be at least this high (useful to prevent
             resizing of the Carousel between slides that are unequal in height). The value 
             is an int denoting a height in terms of the size of the font of the contents (em). 
    """
    def __init__(self, view, css_id, show_indicators=True, interval=5000, pause='hover', wrap=True, keyboard=True, min_height=None):
        super().__init__(view)
        self.carousel_panel = self.add_child(Div(view, css_id=css_id))
        self.carousel_panel.append_class('carousel')
        self.carousel_panel.append_class('slide')
        self.carousel_panel.set_attribute('data-ride', 'carousel')

        self.carousel_panel.set_attribute('data-interval', str(interval))
        pause_option = HTMLAttributeValueOption(pause or 'false', True, constrain_value_to=['hover', 'false'])
        self.carousel_panel.set_attribute('data-pause', pause_option.as_html_snippet())
        self.carousel_panel.set_attribute('data-wrap', 'true' if wrap else 'false')
        self.carousel_panel.set_attribute('data-keyboard', 'true' if keyboard else 'false')
        if min_height:
            style = self.carousel_panel.add_child(HTMLElement(self.view, 'style', children_allowed=True))
            css_id = self.carousel_panel.css_id
            style.add_child(TextNode(self.view, '#%s .carousel-item { min-height: %sem; }' % (css_id, min_height)))

        self.show_indicators = show_indicators
        if self.show_indicators:
            self.indicator_list = self.carousel_panel.add_child(self.create_indicator_list())
        self.inner = self.carousel_panel.add_child(self.create_inner())
        self.slides = []

        self.add_control(previous=True)
        self.add_control()

    def create_inner(self):
        inner = Div(self.view)
        inner.append_class('carousel-inner')
        return inner

    def add_slide(self, widget, caption_widget=None):
        """
        Adds a Slide for the given :class:`~reahl.web.fw.Widget`.

        :param widget: A :class:`~reahl.web.fw.Widget` to display in this slide. (See also :class:`~reahl.web.holder.holder.PlaceholderImage`)
        :keyword caption_widget: The :class:`~reahl.web.fw.Widget` to use as caption for the slide.

        """
        slide = self.inner.add_child(Slide(self.view, widget, caption_widget, len(self.slides)))

        self.slides.append(slide)

        if self.show_indicators:
            self.add_indicator_for(slide)

        return slide

    @property
    def url(self):
        return Url('#%s' % self.carousel_panel.css_id)

    def add_control(self, previous=False):
        control_a = self.carousel_panel.add_child(A(self.view, self.url))
        control_a.append_class('carousel-control-prev' if previous else 'carousel-control-next')
        control_a.set_attribute('role', 'button')
        control_a.set_attribute('data-slide', 'prev' if previous else 'next')

        span_icon = control_a.add_child(Span(self.view))
        span_icon.append_class('carousel-control-%s-icon' % ('prev' if previous else 'next'))
        span_icon.set_attribute('aria-hidden', 'true')

        span_text = control_a.add_child(Span(self.view, text=_('Previous') if previous else _('Next')))
        span_text.append_class('sr-only')

        return control_a

    def create_indicator_list(self):
        indicators = Ol(self.view)
        indicators.append_class('carousel-indicators')
        return indicators

    def add_indicator_for(self, item):
        li = self.indicator_list.add_child(Li(self.view))
        li.set_attribute('data-target', str(self.url))
        li.set_attribute('data-slide-to', '%s' % item.index)
        if item.is_active:
            li.append_class('active')
        return li



# Copyright 2019, 2020, 2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. versionadded:: 5.0

The very basic Bootstrap components (ones that do not warrant their own module)

"""



from reahl.component.exceptions import ProgrammerError
from reahl.component.i18n import Catalogue
from reahl.web.fw import Widget
import reahl.web.ui
from reahl.web.ui import Div, Span, P, A, Url, TextNode, HTMLElement, HTMLAttributeValueOption, ErrorWidget
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.forms import ButtonLayout
from reahl.web.bootstrap.ui import Alert
from reahl.web.bootstrap.grid import ColumnLayout, Container


_ = Catalogue('reahl-web')


class ErrorMessageBox(ErrorWidget):
    def __init__(self, view):
        super().__init__(view)
        alert = self.add_child(Alert(view, _('An error occurred:'), 'danger'))
        alert.add_child(HTMLElement(view, 'hr'))
        alert.add_child(P(view, text=self.error_message))
        a = alert.add_child(A(view, Url(self.error_source_href), description='Ok'))
        a.use_layout(ButtonLayout(style='primary'))


class BootstrapHTMLErrorPage(Widget):
    def __init__(self, view, page_class, *widget_args, **widget_kwargs):
        super().__init__(view)
        page = self.add_child(page_class(view, *widget_args, **widget_kwargs))
        if isinstance(page.layout, PageLayout):
            error_container = page.layout.header
            self.error_box = error_container.add_child(ErrorMessageBox(view))
        else:
            error_container = page.body
            self.error_box = error_container.insert_child(0, ErrorMessageBox(view))

    @classmethod
    def get_widget_bookmark_for_error(cls, error_message, error_source_bookmark):
        return ErrorMessageBox.get_widget_bookmark_for_error(error_message, error_source_bookmark)


class HTML5Page(reahl.web.ui.HTML5Page):
    """A web page that may be used as the page of a web application. It ensures that everything needed by
       the framework (linked CSS and JavaScript, etc) is available on such a page.

       .. admonition:: Styling
       
          Renders as an HTML5 page with customised <head> and an empty <body>.
       
       :param view: (See :class:`reahl.web.fw.Widget`)
       :keyword title: (See :class:`reahl.web.ui.HTML5Page`)
       :keyword css_id: (See :class:`reahl.web.ui.HTMLElement`)
       
    """
    @classmethod
    def get_error_page_factory(cls, *widget_args, **widget_kwargs):
        return BootstrapHTMLErrorPage.factory(cls, *widget_args, **widget_kwargs)
    
    def check_form_related_programmer_errors(self):
        super().check_form_related_programmer_errors()
        self.check_grids_nesting()
        
    def check_grids_nesting(self):
        for widget, parents_set in self.parent_widget_pairs(set([])):
            if isinstance(widget.layout, ColumnLayout):
                if not any(isinstance(parent.layout, Container) for parent in parents_set):
                    raise ProgrammerError(('%s does not have a parent with Layout of type %s.' % (widget, Container))+\
                      ' %s has a ColumnLayout, and thus needs to have an anchestor with a Container Layout.' % widget)



# Copyright 2015-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



from reahl.tofu import scenario, Fixture, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import WidgetTester, XPath

from reahl.web.fw import Url
from reahl.web.bootstrap.ui import P
from reahl.web.bootstrap.tabbedpanel import TabbedPanel, MultiTab, Tab


from reahl.webdev.fixtures import WebServerFixture
from reahl.web_dev.fixtures import WebFixture


@uses(web_fixture=WebFixture)
class   TabbedPanelAjaxFixture(Fixture):

    def new_PopulatedTabbedPanel(self):
        class PopulatedTabbedPanel(TabbedPanel):
            def __init__(self, view):
                super().__init__(view)
                multi_tab = MultiTab(view, 'multitab name', 'multi-main')
                tab1 = Tab(view, 'tab 1 name', 'multi1', P.factory(text='tab 1/1 content'))
                tab2 = Tab(view, 'tab 2 name', 'multi2', P.factory(text='tab 1/2 content'))
                multi_tab.add_tab(tab1)
                multi_tab.add_tab(tab2)
                self.add_tab(multi_tab)

                tab3 = Tab(view, 'tab 3 name', 'tab3', P.factory(text='tab 3 content'))
                self.add_tab(tab3)
                tab4 = Tab(view, 'tab 4 name', 'tab4', P.factory(text='tab 4 content'))
                self.add_tab(tab4)
        return PopulatedTabbedPanel

    def new_wsgi_app(self, enable_js=True):
        return self.web_fixture.new_wsgi_app(enable_js=enable_js, child_factory=self.PopulatedTabbedPanel.factory())

    def tab_is_active(self, tab_name):
        element = self.web_fixture.driver_browser.find_element('//a[contains(text(), "%s")]' % tab_name)
        return 'active' in element.get_attribute('class')

    def tab_contents_equals(self, expected_contents):
        current_contents = self.web_fixture.driver_browser.get_inner_html_for('//div[contains(@class, "active")]')
        return current_contents == expected_contents


@with_fixtures(WebFixture)
def test_basic_rendering(web_fixture):
    """A TabbedPanel consists of a Nav (its tabs) and a Div in which tab contents are displayed."""


    web_fixture.request.query_string = 'tab=tab1'
    tabbed_panel = TabbedPanel(web_fixture.view)
    tabbed_panel.add_tab(Tab(web_fixture.view, 'tab 1 name', 'tab1', P.factory(text='tab 1 content')))

    tester = WidgetTester(tabbed_panel)

    expected_html = \
      '''<ul role="tablist" class="nav nav-tabs reahl-menu">'''\
       '''<li class="nav-item">'''\
       '''<a id="nav_tab_tab1_tab" aria-controls="tab_tab1" aria-selected="true" data-target="#tab_tab1" data-toggle="tab" href="/?tab=tab1" role="tab" class="active nav-link">tab 1 name</a>'''\
       '''</li>'''\
      '''</ul>'''\
      '''<div class="tab-content">'''\
       '''<div id="tab_tab1" aria-labelledby="nav_tab_tab1_tab" role="tabpanel" class="active tab-pane"><p>tab 1 content</p></div>'''\
      '''</div>'''\

    actual = tester.render_html()
    assert actual == expected_html


@with_fixtures(WebFixture)
def test_tabs_with_sub_options(web_fixture):
    """A TabbedPanel can have Tabs that are each composed of multiple sub-options."""

    web_fixture.request.query_string = 'tab=mult2'
    tabbed_panel = TabbedPanel(web_fixture.view)
    multi_tab = MultiTab(web_fixture.view, 'tab 1 name', 'multitab-main')
    multi_tab.add_tab(Tab(web_fixture.view, 'multi tab 1', 'mult1', P.factory(text='tab 1/1 content')))
    multi_tab.add_tab(Tab(web_fixture.view, 'multi tab 2', 'mult2', P.factory(text='tab 1/2 content')))
    tabbed_panel.add_tab(multi_tab)

    tester = WidgetTester(tabbed_panel)

    expected_html = \
     '''<ul role="tablist" class="nav nav-tabs reahl-menu">'''\
     '''<li class="dropdown nav-item">'''\
      '''<a aria-haspopup="true" data-toggle="dropdown" href="/?open_item=tab+1+name&amp;tab=mult2" role="button" class="active dropdown-toggle nav-link reahl-ajaxlink">tab 1 name</a>'''\
      '''<div class="dropdown-menu">'''\
       '''<a id="nav_tab_mult1_tab" aria-controls="tab_mult1" aria-selected="false" data-target="#tab_mult1" data-toggle="tab" href="/?tab=mult1" role="tab" class="dropdown-item">multi tab 1</a>'''\
       '''<a id="nav_tab_mult2_tab" aria-controls="tab_mult2" aria-selected="true" data-target="#tab_mult2" data-toggle="tab" href="/?tab=mult2" role="tab" class="active dropdown-item">multi tab 2</a>'''\
      '''</div>'''\
     '''</li>'''\
    '''</ul>'''\
    '''<div class="tab-content">'''\
     '''<div id="tab_mult1" aria-labelledby="nav_tab_mult1_tab" role="tabpanel" class="tab-pane"><p>tab 1/1 content</p></div>'''\
     '''<div id="tab_mult2" aria-labelledby="nav_tab_mult2_tab" role="tabpanel" class="active tab-pane"><p>tab 1/2 content</p></div>'''\
    '''</div>'''

    actual = tester.render_html()
    assert actual == expected_html


@uses(web_fixture=WebFixture)
class DefaultTabScenarios(Fixture):

    @scenario
    def specified_on_query_string(self):
        self.web_fixture.request.query_string = 'tab=tab2'
        self.expected_contents = '<p>tab 2 content</p>'
        self.tab1_active = False
        self.tab2_active = True

    @scenario
    def defaulted(self):
        self.web_fixture.request.query_string = ''
        self.expected_contents = '<p>tab 1 content</p>'
        self.tab1_active = True
        self.tab2_active = False


@with_fixtures(WebFixture, DefaultTabScenarios)
def test_default_active_tab(web_fixture, default_tab_scenarios):
    """The first tab is active by default (if the active tab is not indicated in the query_string)."""

    tab1 = Tab(web_fixture.view, 'tab 1 name', 'tab1', P.factory(text='tab 1 content'))
    tab2 = Tab(web_fixture.view, 'tab 2 name', 'tab2', P.factory(text='tab 2 content'))

    tabbed_panel = TabbedPanel(web_fixture.view)
    tabbed_panel.add_tab(tab1)
    tabbed_panel.add_tab(tab2)

    [menu_item1, menu_item2] = tabbed_panel.nav.menu_items
    assert menu_item1.is_active == default_tab_scenarios.tab1_active
    assert menu_item2.is_active == default_tab_scenarios.tab2_active

    tester = WidgetTester(tabbed_panel)
    panel_contents = tester.get_html_for('//div[@class="tab-content"]/div[contains(@class, "active")]/*')
    assert panel_contents == default_tab_scenarios.expected_contents


class DefaultMultiTabScenarios(Fixture):
    @scenario
    def specified_on_query_string(self):
        self.query_args = {'tab':'multi2'}
        self.expected_contents = '<p>tab 1/2 content</p>'
        self.multi_tab_active = True
        self.tab1_active = False
        self.tab2_active = True
        self.tab3_active = False

    @scenario
    def defaulted(self):
        self.query_args = {}
        self.expected_contents = '<p>tab 1/1 content</p>'
        self.multi_tab_active = True
        self.tab1_active = True
        self.tab2_active = False
        self.tab3_active = False

    @scenario
    def specified_to_other_tab(self):
        self.query_args = {'tab':'tab3'}
        self.expected_contents = '<p>tab 3 content</p>'
        self.multi_tab_active = False
        self.tab1_active = False
        self.tab2_active = False
        self.tab3_active = True


@with_fixtures(WebFixture, TabbedPanelAjaxFixture, DefaultMultiTabScenarios)
def test_default_active_multi_tab(web_fixture, tabbed_panel_ajax_fixture, default_multi_tab_scenarios):
    """The first item of the first tab is active by default (if the active tab is not indicated in the query_string)."""
    fixture = default_multi_tab_scenarios


    web_fixture.reahl_server.set_app(tabbed_panel_ajax_fixture.wsgi_app)
    url = Url('/')
    url.set_query_from(fixture.query_args)
    web_fixture.driver_browser.open(str(url))

    assert tabbed_panel_ajax_fixture.tab_contents_equals(fixture.expected_contents)

    assert (not fixture.multi_tab_active) or tabbed_panel_ajax_fixture.tab_is_active('multitab name')
    assert (not fixture.tab1_active) or tabbed_panel_ajax_fixture.tab_is_active('tab 1 name')
    assert (not fixture.tab2_active) or tabbed_panel_ajax_fixture.tab_is_active('tab 2 name')
    assert (not fixture.tab3_active) or tabbed_panel_ajax_fixture.tab_is_active('tab 3 name')


@uses(web_fixture=WebFixture, web_server_fixture=WebServerFixture)
class PanelSwitchFixture(Fixture):

    def ensure_disabled_js_files_not_cached(self):
        if self.web_server_fixture.is_instantiated('chrome_driver'):
            assert self.web_fixture.chrome_driver is self.web_fixture.web_driver
            self.web_server_fixture.restart_chrome_session()

    @scenario
    def without_js(self):
        self.enable_js = False

    @scenario
    def with_js(self):
        self.enable_js = True


@with_fixtures(WebFixture, PanelSwitchFixture, TabbedPanelAjaxFixture)
def test_clicking_on_different_tabs_switch(web_fixture, panel_switch_fixture, tabbed_panel_ajax_fixture):
    """Clicking on tabs change the contents that are displayed as well as the active tab."""
    if not panel_switch_fixture.enable_js:
        panel_switch_fixture.ensure_disabled_js_files_not_cached()


    wsgi_app = tabbed_panel_ajax_fixture.new_wsgi_app(enable_js=panel_switch_fixture.enable_js)
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    # Clicking on 3 (a normal tab), changes the current tab
    assert browser.wait_for_not(tabbed_panel_ajax_fixture.tab_is_active, 'tab 3 name')
    assert browser.wait_for_not(tabbed_panel_ajax_fixture.tab_contents_equals, '<p>tab 3 content</p>')

    browser.click(XPath.link().with_text('tab 3 name'))

    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_is_active, 'tab 3 name')
    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_contents_equals, '<p>tab 3 content</p>')

    browser.click(XPath.link().with_text('tab 4 name'))

    assert browser.wait_for_not(tabbed_panel_ajax_fixture.tab_is_active, 'tab 3 name')
    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_is_active, 'tab 4 name')
    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_contents_equals, '<p>tab 4 content</p>')


@with_fixtures(WebFixture, PanelSwitchFixture, TabbedPanelAjaxFixture)
def test_clicking_on_multi_tab(web_fixture, panel_switch_fixture, tabbed_panel_ajax_fixture):
    """Clicking on a multitab just opens and closes its dropdown without affecting the current open tab."""
    if not panel_switch_fixture.enable_js:
        panel_switch_fixture.ensure_disabled_js_files_not_cached()

    wsgi_app = tabbed_panel_ajax_fixture.new_wsgi_app(enable_js=panel_switch_fixture.enable_js)
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    # Make tab 3 the active one
    browser.click(XPath.link().with_text('tab 3 name'))
    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_is_active, 'tab 3 name')
    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_contents_equals, '<p>tab 3 content</p>')

    # Clicking on the multitab toggles the dropdown
    assert browser.wait_for_element_not_visible(XPath.link().with_text('tab 2 name'))
    browser.click(XPath.link().with_text('multitab name'))
    assert browser.wait_for_element_visible(XPath.link().with_text('tab 2 name'))

    # - current active tab not changed
    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_is_active, 'tab 3 name')
    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_contents_equals, '<p>tab 3 content</p>')

    # Clicking on the multitab toggles the dropdown again
    browser.click(XPath.link().with_text('multitab name'))

    # - current active tab not changed
    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_is_active, 'tab 3 name')
    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_contents_equals, '<p>tab 3 content</p>')


@with_fixtures(WebFixture, PanelSwitchFixture, TabbedPanelAjaxFixture)
def test_clicking_on_sub_tab_switches(web_fixture, panel_switch_fixture, tabbed_panel_ajax_fixture):
    """Clicking on a sub tab also changes the contents that are displayed as well as the active tab."""
    if not panel_switch_fixture.enable_js:
        panel_switch_fixture.ensure_disabled_js_files_not_cached()

    wsgi_app = tabbed_panel_ajax_fixture.new_wsgi_app(enable_js=panel_switch_fixture.enable_js)
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    browser.click(XPath.link().with_text('tab 3 name'))
    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_is_active, 'tab 3 name')
    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_contents_equals, '<p>tab 3 content</p>')

    assert browser.wait_for_not(tabbed_panel_ajax_fixture.tab_is_active, 'tab 2 name')
    assert browser.wait_for_not(tabbed_panel_ajax_fixture.tab_contents_equals, '<p>tab 1/2 content</p>')

    browser.click(XPath.link().with_text('multitab name'))
    browser.click(XPath.link().with_text('tab 2 name'))

    # - active status removed from previous
    assert browser.wait_for_not(tabbed_panel_ajax_fixture.tab_is_active, 'tab 3 name')

    # - new status and contents set
    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_is_active, 'tab 2 name')
    assert browser.wait_for(tabbed_panel_ajax_fixture.tab_contents_equals, '<p>tab 1/2 content</p>')

    # Clicking away from the multitab sub-tab removes its active status
    browser.click(XPath.link().with_text('tab 3 name'))
    # tab2 is not active anymore
    assert browser.wait_for_not(tabbed_panel_ajax_fixture.tab_is_active, 'tab 2 name')
    # TODO: cs Remove the next commented code - there used to be a problem with bootstrap4 alpha
    # tab2 is not active anymore
    # if fixture.enable_js:
    #    pass
    #    ### assert None, 'This is a bug in bootstrap v4.0 alpha javascript'
    # else:
    #     vassert( fixture.driver_browser.wait_for_not(fixture.tab_is_active, 'tab 2 name') )


